#ifndef _DDF_HANDLER_H
#define _DDF_HANDLER_H
#include "com_type.h"

#define READ							0
#define WRITE							1
#define GUID_SIZE						24		/*24 bytes*/
#define DDF_REV							0x1
#define LD_TYPE							0		/* No Disk Grouping, LD_Name is ASCII format */
#define	RECORDS_NUM						3		/* 3 LD records per item */
#define BBL_SIZE						5		/* sector */
#define DDF_SIZE						0x20000 /* 64MB */
// For simulator only
//#define DDF_SIZE						128 /* 64k */
#define MBR_SIGNATURE					0xaa55
#define MBR_SIGNATURE_OFFSET			0x01fe
/* PRL */
//#define RAID_0							0
//#define RAID_1							1
//#define RAID_5							5
//#define JBOD							0x0f
//#define RAID_1E							11
//#define RAID_6							0x16
/* SRL */
#define STRIPED							0
#define MIRRORED						1
/* RAID Qualifier */
#define RAID_RLQ						0

//#define DDF_VALID						MV_BIT(0)
//#define DISK_VALID						MV_BIT(1)
//#define DDF_CLEAN						MV_BIT(2)

//VD_State in Virtual Disk Records
//#define DDF_VD_OPTIMAL					0x0
//#define DDF_VD_DEGRADED					0x1
//#define DDF_VD_DELETED					0x2
//#define DDF_VD_MISSING					0x3
//#define DDF_VD_FAILED					0x4
#define DDF_VD_MODE						MV_BIT(0) | MV_BIT(1) | MV_BIT(2)
//#define DDF_VD_MORPHING					MV_BIT(3)
#define DDF_VD_CONSISTENT				MV_BIT(4)

//Following states are saved in vendor scratch space of LD config record
//#define DDF_VD_REBUILD					MV_BIT(0)
//#define DDF_VD_SYNC						MV_BIT(1)
//#define DDF_VD_CHECK					MV_BIT(2)
//#define DDF_VD_INIT						MV_BIT(3)

//Init_State in Virtual Disk Records
#define DDF_VD_CACHE_CLEAN				0x00
#define DDF_VD_CACHE_DIRTY				0x10
#define DDF_VD_NOT_INIT					0x00
#define DDF_VD_QUICK_INITED				0x01
#define DDF_VD_INITED					0x02

//#define PD_STATE_ONLINE					MV_BIT(0)
//#define PD_STATE_FAILED					MV_BIT(1)
//#define PD_STATE_REBUILDING				MV_BIT(2)
//#define PD_STATE_TRANSITION				MV_BIT(3)
//#define PD_STATE_SMART_ERROR			MV_BIT(4)
//#define PD_STATE_READ_ERROR				MV_BIT(5)
//#define PD_STATE_MISSING				MV_BIT(6)

//#define GLOBAL_SPARE_DISK				MV_BIT(2)

#define ANCHOR_DDF_HEADER 				0x00
#define PRIMARY_DDF_HEADER 				0x01
#define SECONDARY_DDF_HEADER 			0x02

#define PRIMARY_HEADER_OFFSET			255			//256 sectors
#define SIGNATURE_DDF_HEADER 			CPU_TO_BIG_ENDIAN_32(0xDE11DE11)
#define SIGNATURE_CONTROLLER_DATA 		CPU_TO_BIG_ENDIAN_32(0xAD111111)
#define SIGNATURE_PD_RECORDS			CPU_TO_BIG_ENDIAN_32(0x22222222)
#define SIGNATURE_PD_DATA 				CPU_TO_BIG_ENDIAN_32(0x33333333)
#define SIGNATURE_LD_RECORDS 			CPU_TO_BIG_ENDIAN_32(0xDDDDDDDD)
#define SIGNATURE_LD_CONFIG_RECORD		CPU_TO_BIG_ENDIAN_32(0xEEEEEEEE)
#define SIGNATURE_SPARE_ASSIGN_RECORD	CPU_TO_BIG_ENDIAN_32(0x55555555)
#define SIGNATURE_VU_CONFIG_RECORD		CPU_TO_BIG_ENDIAN_32(0x88888888)
#define SIGNATURE_VS_LOG				CPU_TO_BIG_ENDIAN_32(0x01DBEEF0)
#define SIGNATURE_BBM_LOG				CPU_TO_BIG_ENDIAN_32(0xABADB10C)

//DDF commands
#define POWER_ON						0
#define UPDATE_LD						1
#define DELETE_LD						2
#define UPDATE_PD						3
#define PLUG_IN							4
#define WIPE_OUT						5
#define TOTAL_DDF_COMMANDS				6			//total of DDF commands

// DDF op-code for CDB
#define ANCHOR_HEADER					MV_BIT(0)
#define PRIMARY_HEADER					MV_BIT(1)		//this covers Primary HEADER, Controller Data, 
#define PD_RECORDS						MV_BIT(2)		//PD Records
#define LD_RECORDS						MV_BIT(3)		//LD Records
#define CONFIG_RECORDS					(CONFIG_RECORDS_0 | CONFIG_RECORDS_1 | CONFIG_RECORDS_2 |CONFIG_RECORDS_3 | CONFIG_RECORDS_4 | CONFIG_RECORDS_5 | \
										 CONFIG_RECORDS_6 | CONFIG_RECORDS_7 | CONFIG_RECORDS_8 |CONFIG_RECORDS_9 | CONFIG_RECORDS_10 | CONFIG_RECORDS_11 | \
										 CONFIG_RECORDS_12 | CONFIG_RECORDS_13 | CONFIG_RECORDS_14 |CONFIG_RECORDS_15 | CONFIG_RECORDS_16)
#define CONFIG_RECORDS_NUM				17
#define CONFIG_RECORDS_0				MV_BIT(4)
#define CONFIG_RECORDS_1				MV_BIT(5)
#define CONFIG_RECORDS_2				MV_BIT(6)
#define CONFIG_RECORDS_3				MV_BIT(7)
#define CONFIG_RECORDS_4				MV_BIT(8)
#define CONFIG_RECORDS_5				MV_BIT(9)
#define CONFIG_RECORDS_6				MV_BIT(10)
#define CONFIG_RECORDS_7				MV_BIT(11)
#define CONFIG_RECORDS_8				MV_BIT(12)
#define CONFIG_RECORDS_9				MV_BIT(13)
#define CONFIG_RECORDS_10				MV_BIT(14)
#define CONFIG_RECORDS_11				MV_BIT(15)
#define CONFIG_RECORDS_12				MV_BIT(16)
#define CONFIG_RECORDS_13				MV_BIT(17)
#define CONFIG_RECORDS_14				MV_BIT(18)
#define CONFIG_RECORDS_15				MV_BIT(19)
#define CONFIG_RECORDS_16				MV_BIT(20)
#define PD_DATA							MV_BIT(21)
#define BBM_LOGS						MV_BIT(22)
#define VENDOR_LOGS						MV_BIT(23)
#define PD_MBR							MV_BIT(24)

#define OP_CODE			(ANCHOR_HEADER | PRIMARY_HEADER | PD_RECORDS | LD_RECORDS | CONFIG_RECORDS | PD_DATA | PD_MBR) // | BBM_LOGS | VENDOR_LOGS //implement it later

#define WORKSPACE_LENGTH				128		//reserve 64MB

#define MAX_PD_ENTRIES					0x3f	//We allocate memory for 64 Physical disks in DDF. 
												//But we only suport 32 disks in our RAID. This is compatibility purpose. 
#define MAX_LD_ENTRIES 					0x3f	//We allocate memory for 64 Logical disks. We only support 32 LD in our RAID.
#define MAX_PARTITIONS					16		//16 extents per disk
#define MAX_PRIMARY_ELEMENT_ENTRIES		64		//MAX 64 disks per LD
#define CONFIG_RECORD_LENGTH			(1+ROUNDING(MAX_PRIMARY_ELEMENT_ENTRIES*(4+8),RAID_SECTOR_SIZE)/RAID_SECTOR_SIZE)	

#define CONTROLLER_DATA_SECTION			(sizeof(DDF_Header)/RAID_SECTOR_SIZE)
#define CONTROLLER_DATA_SECTION_LENGTH	0x00000001	//1 sector

#define PD_RECORDS_SECTION				(CONTROLLER_DATA_SECTION + CONTROLLER_DATA_SECTION_LENGTH)
#define PD_RECORDS_SECTION_LENGTH		0x00000008	//8 sectors

#define LD_RECORDS_SECTION				(PD_RECORDS_SECTION + PD_RECORDS_SECTION_LENGTH)
#define LD_RECORDS_SECTION_LENGTH		0x00000008	//8 sectors

#define CONFIG_RECORDS_SECTION			(LD_RECORDS_SECTION + LD_RECORDS_SECTION_LENGTH)
#define CONFIG_RECORDS_SECTION_LENGTH	(CONFIG_RECORD_LENGTH * (1+MAX_PARTITIONS))	//51 sectors

#define PD_DATA_SECTION					(CONFIG_RECORDS_SECTION + CONFIG_RECORDS_SECTION_LENGTH)
#define PD_DATA_SECTION_LENGTH			0x00000001	//512 bytes

#define BBM_LOG_SECTION					(PD_DATA_SECTION + PD_DATA_SECTION_LENGTH)
#define BBM_LOG_SECTION_LENGTH			0x00000006	//4096 bytes

#define DIAGNOSTIC_SPACE				0xFFFFFFFF
#define DIAGNOSTIC_SPACE_LENGTH			0xFFFFFFFF

#define VENDOR_SPECIFIC_LOGS			(BBM_LOG_SECTION + BBM_LOG_SECTION_LENGTH)
#define VENDOR_SPECIFIC_LOGS_LENGTH		BBL_SIZE	//BBL size	

#ifndef _OS_BIOS
#pragma pack(1)
#endif

typedef struct _DDF_Header {
	MV_U32		Signature;	
	MV_U32 		CRC;
	MV_U8 		DDF_Header_GUID[GUID_SIZE];
	MV_U8  		DDF_rev[8];				
	MV_U32 		Sequence_Number;
	MV_U32 		Timestamp;
	MV_U8  		Open_Flag;						//not implement
	MV_U8  		Foreign_Flag;					//not implement
	MV_U8  		Disk_Grouping;
	MV_U8  		Reserved1[13];
	MV_U8  		Header_Ext[32];					//0xff
	MV_LBA 		Primary_Header_LBA;
	MV_LBA 		Secondary_Header_LBA;
	MV_U8  		Header_Type;
	MV_U8  		Reserved2[3];
	MV_U32		WorkSpace_Length;			//reserve 64MB
	MV_LBA		WorkSpace_LBA;
	MV_U16		Max_PD_Entries;					//support 31 PD
	MV_U16		Max_LD_Entries;					//support 31 LD
	MV_U16		Max_Partitions;					//support 16 extent/PD
	MV_U16		Config_Record_Length;
	MV_U16		Max_Primary_Element_Entries;
	MV_U8		Reserved3[54];
	MV_U32		Controller_Data_Section;
	MV_U32		Controller_Data_Section_Length;
	MV_U32		PD_Records_Section;
	MV_U32		PD_Records_Section_Length;
	MV_U32		LD_Records_Section;
	MV_U32		LD_Records_Section_Length;
	MV_U32		Config_Records_Section;
	MV_U32		Config_Records_Section_Length;
	MV_U32		PD_Data_Section;
	MV_U32		PD_Data_Section_Length;
	MV_U32		BBM_Log_Section;				//it will be implemented in ROC
	MV_U32		BBM_Log_Section_Length;
	MV_U32		Diagnostic_Space;				//not implement
	MV_U32		Diagnostic_Space_Length;		//not implement
	MV_U32		Vendor_Specific_Logs;
	MV_U32		Vendor_Specific_Logs_Length;
	MV_U8		Reserved4[256];
}BUFFER_PACKED DDF_Header, * PDDF_Header;

typedef struct _DDF_Controller_DATA {
	MV_U32 		Signature;                                 
	MV_U32 		CRC;
	MV_U8  		DDF_Controller_GUID[GUID_SIZE];
	MV_U16  	Vendor_ID;
	MV_U16		Device_ID;
	MV_U16		SubVendor_ID;
	MV_U16		SubDevice_ID;
	MV_U8		NID[24];
	MV_U8		BGA_Rate;    
	MV_U8		Controller_Data[447];
}BUFFER_PACKED DDF_Controller_DATA, * PDDF_Controller_DATA;

typedef struct _PD_ENTRY {
	MV_U8 		PD_GUID[GUID_SIZE];      
	MV_U32 		PD_Reference;
	MV_U16 		PD_Type;
	MV_U16 		PD_State;
	MV_U64 		Configured_Size;
	MV_U8 		Path_Info[18];
	MV_U8 		Reserved[6];
}BUFFER_PACKED PD_ENTRY, * PPD_ENTRY;						//64 bytes

typedef struct _DDF_PD_Records {
	MV_U32 		Signature;                                 
	MV_U32 		CRC;
	MV_U16  	Populated_PDEs;
	MV_U16  	Max_PD_Supported;
	MV_U8  		Reserved[52];
	PD_ENTRY	PD_Entries[MAX_PD_ENTRIES];	//64*63 + 64 bytes
}BUFFER_PACKED DDF_PD_Records, * PDDF_PD_Records;

typedef struct _Prim_Header{
	DDF_Header				header;
	DDF_Controller_DATA	controller_data;
} Prim_Header, * PPrim_Header;

typedef struct _LD_Entry{
	MV_U8 		LD_GUID[GUID_SIZE];   
	MV_U16 		LD_Number;
	MV_U16		Reserved0;
	MV_U32 		LD_Type;
	MV_U8 		LD_State;
	MV_U8 		Init_State;
	MV_U8 		Reserved1[14];
	MV_U8 		LD_Name[LD_MAX_NAME_LENGTH];
}BUFFER_PACKED LD_Entry, *PLD_Entry;		/* 64 bytes */

typedef struct _DDF_LD_Records{
	MV_U32 		Signature;                                 
	MV_U32 		CRC;
	MV_U16  	Populated_VDEs;
	MV_U16  	Max_LD_Supported;
	MV_U8  		Reserved[52];
	LD_Entry  	LD_Entries[MAX_LD_ENTRIES];	//64*63 +64 bytes
}BUFFER_PACKED DDF_LD_Records, *PDDF_LD_Records;

typedef struct _DDF_LD_Config_Records{
	MV_U32		Signature;                                 
	MV_U32		CRC;
	MV_U8 		LD_GUID[GUID_SIZE]; 
	MV_U32 		Timestamp;
	MV_U32 		Sequence_Number;
	MV_U8  		Reserved1[24];
	MV_U16 		Primary_Element_Count;
	MV_U8 		Stripe_Size;
	MV_U8 		PRL;
	MV_U8 		RLQ;
	MV_U8 		Secondary_Element_Count;
	MV_U8 		Secondary_Element_Seq;
	MV_U8 		SRL;
	MV_U64 		Block_Count;
	MV_U64	 	Size;
	MV_U8 		Reserved2[8];
	MV_U8 		Associated_Spares[32];
	MV_U8 		Cache_Policy[8];
	MV_U8 		BG_Rate;
	MV_U8 		Reserved3[3]; 	
	MV_U8 		Reserved4[52];
	MV_U8 		Reserved5[192];
	MV_U8 		V0[32];
	MV_U8 		V1[32];
	MV_U8 		V2[16];
	MV_U8 		V3[16];
	MV_U8		MV_LD_State;			//we use scratch space for extra LD state
	MV_U8		MV_Rebuild_Policy;
	MV_U8		BGA_State;
	MV_U8		MV_LD_Id;				//in order to make LD ID consistent after reboot, we need store LD id.
	MV_LBA		Watermark;
	MV_U16		MV_Target_ID;
	MV_U8 		MV_Num_Parity_Disk;		// For RAID6 and RAID60 only
	MV_U8 		Scratch_Space[17];
	MV_U32		Physical_Disk_Sequence[MAX_PRIMARY_ELEMENT_ENTRIES];
	MV_U64		Starting_Block[MAX_PRIMARY_ELEMENT_ENTRIES];
}BUFFER_PACKED DDF_LD_Config_Records, *PDDF_LD_Config_Records;

typedef struct _Spare_Entry{
	MV_U8 	VD_GUID[GUID_SIZE]; 
	MV_U16 	Secondary_Element;
	MV_U8  	Reserved[6];
}Spare_Entry, *PSpare_Entry;

typedef struct _DDF_Spare_Record{
	MV_U32 		Signature;                                 
	MV_U32 		CRC;
	MV_U32 		Timestamp; 
	MV_U8  		Reserved1[7];
	MV_U8 		Spare_Type;
	MV_U16 		Populated_SAEs;
	MV_U16 		Max_SAE_Supported;
	MV_U8  		Reserved2[8];
	Spare_Entry	Spare_Assignment_Entries[1];
}BUFFER_PACKED DDF_Spare_Record, *PDDF_Spare_Record;

typedef struct _DDF_PD_Data{
	MV_U32 	Signature;                                 
	MV_U32 	CRC;
	MV_U8 	PD_GUID[GUID_SIZE]; 
	MV_U32 	PD_Reference;
	MV_U8	Forced_Ref_Flag;
	MV_U8	Forced_PD_GUID_Flag;
	MV_U8	Scratch_Space[32];
	MV_U8	Reserved[442];
}BUFFER_PACKED DDF_PD_Data, *PDDF_PD_Data;

typedef struct _Mapped_Block_Entries{
	MV_LBA	Defective_Block_Start;                                 
	MV_U32 	Spare_Block_Offset;
	MV_U16 	Remapped_Count; 
	MV_U16 	Reserved;
}BUFFER_PACKED Mapped_Block_Entries, *PMapped_Block_Entries;

typedef struct _DDF_BBM_Log{
	MV_U32	Signature;                                 
	MV_U32 	CRC;
	MV_U16 	Entry_Count; 
	MV_U32 	Reserved_Spare_Block_Count;
	MV_U8 	Reserved[10];
	MV_U8 	First_Spare_LBA[8];
	MV_U8	Mapped_Block_Entries[4064];
}BUFFER_PACKED DDF_BBM_Log, *PDDF_BBM_Log;

#ifndef _OS_BIOS
#pragma pack()
#endif

void DDF_HandleRequest (PRAID_Core	pRaidCore, PMV_Request	 	pRequest);

MV_BOOLEAN GenerateRequest(PRAID_Core pRaidCore, PMV_Request pRequest);

MV_BOOLEAN FillAnchorHeader(PDDF_Header			pDDFHeader, PPD_Disk			pDisk);

MV_BOOLEAN FillPrimHeader(
	IN OUT PRAID_Core		pRaidCore,	
	IN  PPrim_Header		pPrim_Header
);

MV_BOOLEAN FillPDRecords(
	IN  PDDF_PD_Records		pPDRecords,
	IN OUT PPD_Disk			pDisk
);

MV_BOOLEAN FillLDRecords(
	IN  PDDF_LD_Records		pLDRecords,
	IN  MV_U8				pdIndex,
	IN OUT PRAID_Core		pRaidCore			
);

MV_BOOLEAN FillConfigRecords(
	IN  PDDF_LD_Config_Records		pConfigRecords,
	IN  PPD_Disk					pDisk,
	IN  MV_U8						PDIndex,
	IN  PRAID_Core					pRaidCore			
);

MV_BOOLEAN FillPDData(
	IN  PDDF_PD_Data		pPDData,
	IN  PPD_Disk			pDisk
);

MV_BOOLEAN FillBBMData(
	IN  PDDF_BBM_Log		pBBMLog,
	IN  PRAID_Core			pRaidCore			
);

MV_BOOLEAN FillMBR(
	IN  MV_PU8		sBuf,
	IN  PPD_Disk	pDisk
);

PScratch_Buffer GenerateAnchor(
	IN	PPD_Disk			pDisk,
	IN  PDisk_Context		pDiskContext,
	IN  PRAID_Core			pRaidCore	
);

PScratch_Buffer GeneratePrimHeader(
	IN	PPD_Disk			pDisk,
	IN  PDisk_Context		pDiskContext,	
	IN  PRAID_Core			pRaidCore	
);

PScratch_Buffer GeneratePDRecords(
	IN  PDisk_Context		pDiskContext,
	IN  PPD_Disk			pDisk,
	IN  PRAID_Core			pRaidCore
);

PScratch_Buffer GenerateLDRecords(
	IN  PDisk_Context		pDiskContext,
	IN	PPD_Disk			pDisk,
	IN  PRAID_Core			pRaidCore								  
);

PScratch_Buffer GenerateConfigRecords(
	IN  PDisk_Context		pDiskContext,
	IN	MV_U32				opCode,
	IN	PPD_Disk			pDisk,
	IN  PRAID_Core			pRaidCore								  
);

PScratch_Buffer GeneratePDData(
	IN  PDisk_Context		pDiskContext,
	IN	PPD_Disk			pDisk,
	IN  PRAID_Core			pRaidCore	
);

PScratch_Buffer GenerateBBMLog(
	IN  PDisk_Context		pDiskContext,
	IN	PPD_Disk			pDisk,
	IN  PRAID_Core			pRaidCore
);

PScratch_Buffer GenerateMBR(
	IN  PDisk_Context		pDiskContext,
	IN  PRAID_Core			pRaidCore	
);

void InitHeader(
	IN MV_U8				type,
	IN PPD_Disk				pDisk,
	IN OUT PDDF_Header		pDDFHeader,
	IN  PRAID_Core			pRaidCore
);

void UpdateContext(
	IN MV_U32				opCode,
	IN PScratch_Buffer		sBuf,
	IN OUT PDisk_Context	pDiskContext
);

MV_U8 FindBitNum(MV_U32 value);

MV_BOOLEAN IsRedundantRAID(MV_U8 level);

void DDF_Completion(
	IN OUT  PRAID_Core		pRaidCore,
	IN	PMV_Request			pRequest
);

void GenerateDDFGUID(
	IN MV_PU8			pGUID,
	IN  PRAID_Core		pRaidCore
);

void releaseResource(
	IN  PDisk_Context		pDiskContext,		
	IN	PMV_Request			pRequest,
	IN  PRAID_Core			pRaidCore	
);

void BigEndian(
	IN OUT MV_PU8	buf,
	IN MV_U8		len
); 

MV_BOOLEAN FindLDMember(
	IN PRAID_Core		pRaidCore,	
	IN PRAID_Configure	pRaidConf,
	IN PPD_Disk			pDisk
);

MV_U8 FindUpdateInfo(
	IN PRAID_Core		pRaidCore,	
	IN PMV_Request		pRequest,
	IN MV_PU8			pdIndx
);

MV_U8 FindPDInfo(
	IN PRAID_Core		pRaidCore,	
	IN MV_PU8			pdIndx
);
#endif /* _DDF_HANDLER_H */
