#ifndef __RAID_CONSTANT_H__
#define __RAID_CONSTANT_H__

#define RAID_MAX_NUM_DISKS					8

#ifndef _OS_BIOS
	#ifndef SUPPORT_RAID6
		#define RAID_MAX_MV_REQUEST_PER_IO			8
		#define RAID_MAX_XOR_REQUEST_PER_IO			1
		#define RAID_MAX_STRIPE_NODE_PER_IO			4
		#define RAID_MAX_MAP_NODE_PER_STRIPE		4
		#define RAID_MAX_SEQUENCE_NODE_PER_STRIPE	4
		#define RAID_MAX_SCRATCH_BUF_NUM			3
	#else
		#define RAID_MAX_MV_REQUEST_PER_IO			8
		#define RAID_MAX_STRIPE_NODE_PER_IO			4
		#define RAID_MAX_XOR_REQUEST_PER_IO			RAID_MAX_STRIPE_NODE_PER_IO
		#define RAID_MAX_MAP_NODE_PER_STRIPE		4
		#define RAID_MAX_SEQUENCE_NODE_PER_STRIPE	4
		#define RAID_MAX_SCRATCH_BUF_NUM			64	//TBD: Enough
	#endif
#else	/* _OS_BIOS */
	#define RAID_MAX_MV_REQUEST_PER_IO			1
	#define RAID_MAX_XOR_REQUEST_PER_IO			1
	#define RAID_MAX_STRIPE_NODE_PER_IO			2
	#define RAID_MAX_MAP_NODE_PER_STRIPE		2
	#define RAID_MAX_SEQUENCE_NODE_PER_STRIPE	2
	#define RAID_MAX_SCRATCH_BUF_NUM			1
#endif	/* _OS_BIOS */

#define RAID_MAX_BLOCK_DISK					32
#define RAID_MAX_BLOCK_PER_PD				8
#define RAID_MIN_BLOCK_SIZE					2048	/*1MB*/

#define RAID_SCRATCH_BUF_SIZE	0x1000		// 4KB
#define RAID_SECTOR_SIZE	512
#define RAID_QUICK_INIT_SIZE	128		//64KB

#define RAID_HOST_SERVICE_COUNT	0xFB

#define RAID_STATE_LOCK_ACQUIRE			0
#define RAID_STATE_STRIPE_GENERATE		1
#define RAID_STATE_DETERMINE_POLICY		2
#define RAID_STATE_PDMAP_GENERATE		3
#define RAID_STATE_SEQUENCE_GENERATE	4
#define RAID_STATE_SEQUENCE_EXECUTE		5
#define RAID_STATE_COMPLETE				6

#define RAID_NODE_TYPE_STRIPE	0
#define RAID_NODE_TYPE_MAP		1
#define RAID_NODE_TYPE_SEQUENCE	2

#define RAID_SEQ_TYPE_NONE			0
#define RAID_SEQ_TYPE_DISK_ACCESS	1
#define RAID_SEQ_TYPE_XOR			2

#define RAID_CONFIGURE_INVALID		0
#define RAID_CONFIGURE_VALID		1
#define RAID_CONFIGURE_DELETING		2

#define RAID_READ_OP	0x28
#define RAID_WRITE_OP	0x2a
#define RAID_VERIFY_OP	0x2f
#define RAID_REBUILD_OP	0xD0
#define RAID_INIT_OP	0xD1
#define RAID_SYNC_OP	0xD2

#define RAID_SUCCESS	0
#define RAID_FAILED		1

#define RAID_AUTO_REBUILD_ON  0x01
#define RAID_BGA_ACTION_RATIO	10

#define RAID_FLAG_LD_IS_FULL_INITED			MV_BIT(1)	/* //2 1 quick, //0 - not initialized */
#define RAID_FLAG_BGA_STRIPE_NOT_COMPLETED	MV_BIT(8)
#define RAID_FLAG_BGA_TERMINATE_IMMEDIATE	MV_BIT(9)
#define RAID_FLAG_BGA_NECESSARY_CHECK		MV_BIT(10)
#define RAID_FLAG_WRITE_UPDATE_DDF			MV_BIT(11)

#define RAID_RDISK_DDF_WIPED				MV_BIT(1)

#define RAID_DISK_TYPE_BLOCK		0
#define RAID_DISK_TYPE_SUBLD		1

#define RAID_1GB_IN_SECTOR	0x1dcd65L	// 1000*1000*1000/512
#define RAID_10GB_IN_SECTOR 0x12a05f2	//10*1000*1000*1000/512

#define RAID_WIPEOUT_FORCE	1

#define DDF_BUSY			1
#define MV_GUID_SIZE		8
#endif
