#ifndef __RAID_STRUCTURE_H__
#define __RAID_STRUCTURE_H__

#include "com_list.h"
#include "com_error.h"
#include "raid_constant.h"
#ifdef SUPPORT_RAID6
#include "math_matrix.h"
#endif

struct _RAID_Core;
struct _RAID_Layout;
struct _RAID_Request;
struct _Stripe_Info;

typedef struct _Block_Disk {
	MV_LBA	LBA_Offset;
	MV_LBA	Block_Disk_Size;
	MV_U16	Device_ID;
//	MV_U8	PD_State;	//redundant with State in PD_Disk
	MV_U8	Flags;
	MV_U8	PD_Index;
	MV_U8	Block_Disk_ID;
	MV_U8	Reserved[3];
} Block_Disk, *PBlock_Disk;

typedef struct _RAID_Process_Function {
	void (*getLockRange)(struct _RAID_Core*, 
						 struct _RAID_Layout*, 
						 struct _RAID_Request*,
						 MV_LBA*, 
						 MV_LBA*);
	void (*determinePolicy)(struct _RAID_Core*, 
						struct _RAID_Layout*, 
						struct _RAID_Request*,
						struct _Stripe_Info*,
						MV_LBA, 
						MV_LBA);
	MV_U8 (*pdmap)(struct _RAID_Core*, 
				   struct _RAID_Layout*,
				   struct _RAID_Request*,
				   struct _Stripe_Info*, 
				   MV_PU32, 
				   MV_LBA, 
				   MV_LBA);
	MV_U8 (*sequenceGenerater)(struct _RAID_Core*, 
							   struct _RAID_Layout*, 
							   struct _RAID_Request*, 
							   struct _Stripe_Info*,
							   PMV_SG_Entry*,
							   MV_U32*);
} RAID_Process_Function;

typedef struct _RAID_Disk
{
	MV_U8 Block_Disk_Index;
	MV_U16 Block_Disk_State;
	MV_U8 Valid;
} RAID_Disk, *PRAID_Disk;

typedef struct _RAID_BGA_Descriptor
{
	MV_U8 BGA_State;
	MV_U8 BGA_Percentage;
	MV_U8 BGA_Action_Ratio;
	MV_U8 Reserved[1];
	PMV_Request BGA_Control_Req;
	PMV_Request BGA_Running_Req;
	MV_LBA Watermark;
	MV_LBA BGA_End_LBA;
} RAID_BGA_Descriptor, *PRAID_BGA_Descriptor;

typedef struct _RAID_Layout
{
	MV_LBA					LD_Size;		/* Logical Drive size in sector */
	MV_U8					RAID_Level;
	MV_U8					SRL;
	MV_U8					RLQ;
	MV_U8					Stripe_Unit_Power;
	MV_U16					Stripe_Unit_Size;
	MV_U8					Num_Raid_Disk;
	MV_U8					Num_Data_Disk;
	MV_U32					Stripe_Size;	/* Stripe size in sector */
	RAID_Disk				RDisk[RAID_MAX_NUM_DISKS];
	RAID_Process_Function	Raid_Funcs;

#ifdef SUPPORT_RAID6
	MV_U32					Failed_RDisk;			/* It's a bitmap for failed RDisk */
	MV_U8					Num_Parity_Disk;		/* Num_Parity_Disk=Num_Raid_Disk-Num_Data_Disk */
	MV_BOOLEAN				Pattern_Left;			/* default=MV_TRUE */
	MV_BOOLEAN				Pattern_Symmetric;		/* default=MV_TRUE: So left-symmetric */
	MV_U8					Pattern_Parity_Shift;	/* default=1. Shift parity how many blocks. 
													Pattern_Left=TRUE, left shift. Otherwise, right shift.*/
	MV_U8					Reserved1[4];
	Matrix					VD_Matrix;
#endif
} RAID_Layout, *PRAID_Layout;

typedef struct _RAID_Configure
{
	MV_U8				LD_Name[LD_MAX_NAME_LENGTH];
	MV_U8				LD_GUID[MV_GUID_SIZE];
	MV_U32				Generation;
	RAID_BGA_Descriptor	BGA_Descriptor;
	RAID_Layout			Original_Layout;
#ifndef BIOS_NOT_SUPPORT
	RAID_Layout			Migrate_Layout;
#endif
	PRAID_Layout		pCurrent_Layout;
	List_Head			Queue_Lock;
	MV_U8				LD_State;
	MV_U8				MV_LD_State;
	MV_U16				MV_Target_ID;
	MV_U16				LD_Flag;
	MV_U8				BGA_Rate;
	MV_U8				Cache_Policy;
	MV_U8				Rebuild_Policy;
	MV_U8				Valid;
	MV_U16				Core_Device_Id;	
	MV_U8				Bypass_Flag;			// if free disk, bypass requests to core		
	MV_U8				Reserved[3];
} RAID_Configure, *PRAID_Configure;

typedef struct _Scratch_Buffer
{
	struct _Scratch_Buffer	*next;
	MV_PVOID				pSBuf;
	MV_PHYSICAL_ADDR		Physical_Addr;
	MV_U32					Buf_Size;
} Scratch_Buffer, *PScratch_Buffer;

typedef struct _Map_Info
{
	struct _Map_Info	*next;
	MV_U32				Data_Offset;	/* From the start of the MV_Request */
	MV_LBA				Start_LBA;		/* RDisk start LBA. Offset of the RDisk is not counted yet. */
	MV_U32				Sector_Count;	/* Access size */
	MV_U8				RDisk_ID;
	MV_U8				Reserved[3];
#ifdef SUPPORT_RAID6
	PScratch_Buffer		pScratch_Buffer;
#endif
} Map_Info, *PMap_Info;

typedef struct _Sequence_Info
{
	struct _Sequence_Info	*next;
	struct _Stripe_Info		*pStripe_Node;
	struct _RAID_Request	*pRaid_Request;
	List_Head				Queue_Execute_Request;
	List_Head				Queue_Complete_Request;
	MV_U8					Num_Requests;
	MV_U8					Seq_Type;
	MV_U8					Xor_Result;
	MV_U8					Reserved[1];
	void (*Function)(struct _RAID_Core*, struct _Sequence_Info*);
} Sequence_Info, *PSequence_Info;

typedef struct _Stripe_Info
{
	struct _Stripe_Info	*next;
	PMap_Info			pMap_Node_Head;
#ifdef SUPPORT_RAID6
	PMap_Info			pMap_Extra_Node_Head;
	PMap_Info			pMap_Parity_Node_Head;
	MV_U32				Policy;					/* Determine the handling policy for the whole state machine. */
#endif
	PSequence_Info		pSequence_Node_Head;
} Stripe_Info, *PStripe_Info;

typedef struct _Lock_Info
{
	List_Head				Queue_Pointer;
	MV_LBA					Lock_Start_LBA;
	MV_LBA					Lock_End_LBA;
	struct _RAID_Request	*pRaid_Request;
} Lock_Info, *PLock_Info;

typedef struct _RAID_Request
{
	struct _RAID_Request	*next;
	MV_U16					Num_Stripes;
	MV_U16					Num_Complete_Stripes;
	MV_U8					Request_State;
	MV_U8					Complete_Status;
//	MV_U8					Request_Flag;
	MV_U8					Reserved[2];
#ifdef SUPPORT_RAID6
	MV_U32					Failed_RDisk;		/* It's a bitmap for failed RDisk */	//TBD: May need move to somewhere else
	MV_U32					Handling_Policy;	/* Force RAID core to use some special policy. */
#endif
	PMV_Request				pMv_Request;
//	struct _RAID_Core		*pRaid_Core;
	PStripe_Info			pStripe_Node_Head;
#ifndef BIOS_NOT_SUPPORT
	PLock_Info				pLock_Node;
#endif
} RAID_Request, *PRAID_Request;

typedef struct _Node_Pool
{
	MV_PVOID	pPool_Head;
	MV_U32		Free_Count;
} Node_Pool, *PNode_Pool;

typedef struct _Global_Setting
{
	MV_U8	Alarm_On;
	MV_U8	Auto_Rebuild_On;
	MV_U8	BGA_Rate;
	MV_U8	SMART_Status_Timer_Handle;  // if equals to NO_CURRENT_TIMER, don't get SMART status
} Global_Setting, *PGlobal_Setting;

typedef struct _Uncached_Memory
{
	MV_PVOID			pVirtual_Address;
	MV_PHYSICAL_ADDR	Physical_Address;
} Uncached_Memory;

/* From Qun Zhao */
typedef struct _PD_Disk {
	MV_U16	Device_ID;
	MV_U32	Ref_Num;
	MV_U16	State;
	MV_U16	Type;
	MV_LBA	Max_LBA;
	MV_U8	Valid;
	MV_U8	Reserved[1];
	MV_U32	PD_Records_Section;
	MV_U32	PD_Records_Section_Length;
	MV_U32	LD_Records_Section;
	MV_U32	LD_Records_Section_Length;
	MV_U32	Config_Records_Section;
	MV_U32	Config_Records_Section_Length;
	MV_U32	PD_Data_Section;
	MV_U32	PD_Data_Section_Length;
	MV_U32	BBM_Log_Section;
	MV_U32	BBM_Log_Section_Length;
	MV_U32	Vendor_Logs_Section;
	MV_U32	Vendor_Logs_Section_Length;
	MV_LBA	Primary_Header_LBA;
	MV_U16	Max_PD_Entries;
	MV_U16	Config_Record_Length;
	MV_U16	Max_Prim_Element_Entries;
	MV_U16	Max_LD_Entries;
} PD_Disk, *PPD_Disk;

typedef struct _Disk_Context{
	MV_U32			OpCode;
	MV_U32			Processing_OpCode;
	MV_U8			WR;	
	MV_U8			PD_Index;
	MV_U8			Num_Disk;
	MV_U8			LD_Id;
	PScratch_Buffer	sBuf;
} Disk_Context, *PDisk_Context;

typedef struct _RAID_Core
{
	MV_U8			IO_Count;
	MV_U8			Reduced_Mode;
	MV_U16			BGA_Waiting_Count;
#ifndef _OS_BIOS
	MV_U32			DDF_Busy;
#endif
	Global_Setting	Global_Setting_Info;
	Uncached_Memory		Uncached_Mem_Info;
	RAID_Configure	Raid_conf[MAX_LD_SUPPORTED];		// note: after mixing non-raid and raid, this now 
														//		 includes both disks with LD and without
	Block_Disk		Block_Disk_Info[RAID_MAX_BLOCK_DISK];
	PD_Disk			PD_Disk[MAX_HD_SUPPORTED];
	Disk_Context	DiskContext[MAX_HD_SUPPORTED];
	List_Head		Queue_OS_Req;
	List_Head		Queue_BGA_Req;
	List_Head		Queue_DDF_Req;
	List_Head		Queue_MvRequest_Pool;
	List_Head		Queue_XorRequest_Pool;
	List_Head		Queue_LockNode_Pool;
	PRAID_Request	pRAID_Request_Pool_Head;
	Node_Pool		Stripe_Node_Pool;
	Node_Pool		Map_Node_Pool;
	Node_Pool		Sequence_Node_Pool;
	Node_Pool		Scratch_Buf_Pool;
} RAID_Core, *PRAID_Core;

typedef void (*RAID_Management_Command_Handler)(PRAID_Core, PMV_Request);

#endif
