#ifndef __RAID_SUPPORT_H__
#define __RAID_SUPPORT_H__

#define RAID_DBG_INFO(_x_)
#define RAID_ERROR(_x_)

#if 0
#define RAID_DBG_INFO(_x_)	do {printf("RAID INFO: ");	printf _x_; } while (0)
//#define RAID_ERROR(_x_)		do {printf("RAID ERROR: ");	printf _x_; } while (0)
#define RAID_ERROR(_x_)		do { MV_DPRINT(("RAID ERROR: ")); MV_DPRINT(_x_);} while (0)
#endif

#define RAID_MOVE_POOL_HEAD(ptr, type, count) {	\
	type *p_tmp = NULL;	\
	MV_DASSERT( (count>0) && (ptr!=NULL) );	\
	while (count) {	\
		p_tmp = (type *)ptr;	\
		ptr = (MV_PVOID)(p_tmp->next);	\
		count--;	\
	}	\
	if (p_tmp) p_tmp->next = NULL;	\
}

#define RAID_INSERT_NODE(phead, type, pnode) { \
	((type *)pnode)->next = (type *)phead;	\
	phead = pnode;	\
}

#define RAID_SET_SEQ_NODE(pSeq, pReq, pStripe, func, type) {	\
	pSeq->Function = func;	\
	pSeq->pRaid_Request = pReq;	\
	pSeq->pStripe_Node = pStripe;	\
	pSeq->Seq_Type = type;	\
}

#define RAID_SET_SEQ_FUNC(pSeq, func) {	\
	pSeq->Function = func;	\
}

#define RAID_IS_VALID(flag)			( (flag & RAID_CONFIGURE_VALID)==RAID_CONFIGURE_VALID )

#define RAID_IS_LD_AUTOREBUILD_ON(policy)	( policy==RAID_AUTO_REBUILD_ON )
#define RAID_IS_LD_REDUNDANT(level) ( level==LD_MODE_RAID1 )
#define RAID_IS_LD_FUNCTIONAL(state)	( state==LD_STATUS_FUNCTIONAL )
#define RAID_IS_LD_DEGRADED(state)	( state==LD_STATUS_DEGRADE )
#define RAID_IS_LD_WORKING(state)	( (state!=LD_STATUS_DELETED) &&	\
									  (state!=LD_STATUS_MISSING) &&	\
									  (state!=LD_STATUS_OFFLINE) )
#define RAID_IS_LD_FULL_INITED(flag)	( (flag & RAID_FLAG_LD_IS_FULL_INITED)==RAID_FLAG_LD_IS_FULL_INITED )
#define RAID_IS_BGA_TERMINATE_IMMEDIATE(flag) ( (flag & RAID_FLAG_BGA_TERMINATE_IMMEDIATE)==RAID_FLAG_BGA_TERMINATE_IMMEDIATE )

#define RAID_IS_DISK_VALID(valid)		( (valid & PD_DISK_VALID)==PD_DISK_VALID )
#define RAID_IS_DISK_DDF_VALID(valid)	( (valid & PD_DDF_VALID)==PD_DDF_VALID )
#define RAID_IS_DISK_MBR_VALID(valid)	( (valid & PD_MBR_VALID)==PD_MBR_VALID )
#define RAID_IS_DISK_SPARE(type)		( (type & GLOBAL_SPARE_DISK)==GLOBAL_SPARE_DISK )
#define RAID_IS_DISK_REBUILD(state)		( state & PD_STATE_REBUILDING )
#define RAID_IS_DISK_FAILED(state)		( state & PD_STATE_FAILED )
#define RAID_IS_DISK_ASSIGNED(flags)	( (flags & BLOCK_ASSIGNED)==BLOCK_ASSIGNED )
#define RAID_IS_DISK_ATAPI(HDtype)		( (HDtype & HD_TYPE_ATAPI)==HD_TYPE_ATAPI )
#define RAID_IS_RDISK_WIPED(valid)		( (valid & RAID_RDISK_DDF_WIPED)==RAID_RDISK_DDF_WIPED )

#define RAID_CAN_BGA_AUTO_START(state)	( (state != LD_BGA_STATE_ABORTED) && (state != LD_BGA_STATE_PAUSED) )
#define RAID_IS_BGA_RUNNING(state)		( state==LD_BGA_STATE_RUNNING )
#define RAID_IS_BGA_ABORTED(state)		( state==LD_BGA_STATE_ABORTED )

#define RAID_HAS_BGA_IN_LD(state)		( state!=LD_BGA_STATE_NONE )
#define RAID_HAS_LD_REBUILD(state)		( (state & LD_BGA_REBUILD)==LD_BGA_REBUILD )
#define RAID_HAS_LD_BKINIT(state)		( (state & LD_BGA_INIT_BACK)==LD_BGA_INIT_BACK )
#define RAID_HAS_LD_SYNC_CHECK(state)	( (state & LD_BGA_CONSISTENCY_CHECK)==LD_BGA_CONSISTENCY_CHECK )
#define RAID_HAS_LD_SYNC_FIX(state)		( (state & LD_BGA_CONSISTENCY_FIX)==LD_BGA_CONSISTENCY_FIX )
#define RAID_HAS_LD_MIGRATE(state)		( (state & LD_BGA_MIGRATION)==LD_BGA_MIGRATION )

#define RAID_IS_BGA_CHECK_NEEDED(flag)	( (flag & RAID_FLAG_BGA_NECESSARY_CHECK)==RAID_FLAG_BGA_NECESSARY_CHECK )

#define RAID_SET_DISK_FREE(flags)	do { flags &= ~BLOCK_ASSIGNED; } while (0)

#define RAID_IS_EVEN(num)	( ((num) & 0x01)==0 )

#define RAID_GET_MIRROR(i)		( RAID_IS_EVEN(i)? (i+1) : (i-1) )

#define RAID_IS_RETRIED_MVREQ(flag)		( (flag & REQ_FLAG_RETRY)==REQ_FLAG_RETRY )
#define RAID_SET_MVREQ_RETRY(flag)		do { flag |= REQ_FLAG_RETRY; } while (0)
#define RAID_CLEAR_MVREQ_RETRY(flag)	do { flag &= ~REQ_FLAG_RETRY; } while (0)

#define RAID_IS_RWV_REQ(cdb0)	( (cdb0==RAID_READ_OP) ||	\
								  (cdb0==RAID_WRITE_OP) ||	\
								  (cdb0==SCSI_CMD_VERIFY_10) )
#define RAID_IS_BGA_REQ(cdb0)	( (cdb0==RAID_REBUILD_OP) ||	\
								  (cdb0==RAID_INIT_OP) ||	\
								  (cdb0==RAID_SYNC_OP) )	
#define RAID_IS_DDF_BUSY(busyFlag)	( busyFlag==DDF_BUSY )

#define RAID_IS_SEQ_ACCESS_DISK(seqType)	( seqType==RAID_SEQ_TYPE_DISK_ACCESS )
#define RAID_IS_SEQ_XOR(seqType)			( seqType==RAID_SEQ_TYPE_XOR )

#define RAID_IS_BYPASS_DEVICE(bypassFlag)	( bypassFlag==MV_TRUE )
#endif
