#if !defined(__RAID_UTIL_H__)
#define __RAID_UTIL_H__

#include "raid_res_mgmt.h"

#define RAID_GET_COUNT(length, size)	((length) / (size) + 1)
#define RAID_GET_COUNT_U64(length64, size)	((length64).value / (size) + 1)
#define RAID_GET_ID(LBA, sz)	((LBA).value / (sz))
#define RAID_GET_ID_BY_POWER(LBA, power)	((LBA).value >> (power))

/* [start, end). Using these two API, start LBA is included in this stripe but end LBA is not. */
MV_LBA RAID_GET_STRIPE_START_LBA(MV_LBA startLBA, MV_U32 ssz);
MV_LBA RAID_GET_STRIPE_END_LBA(MV_LBA startLBA, MV_U32 ssz);
MV_LBA RAID_GET_UNIT_END_LBA(MV_U32 sunum, MV_U8 susp);

MV_VOID GenerateGUID(PRAID_Configure pRaidConf, MV_U8 ldIdx);
MV_U8 raid_get_available_blk_index(PRAID_Core pRaidCore);
MV_U8 raid_get_available_pd_index(PRAID_Core pRaidCore);
MV_U8 raid_get_available_ld_index(PRAID_Core pRaidCore);

PRAID_Configure raid_get_config_from_ld_id(PRAID_Core pRaidCore, MV_U8 ldId);
PRAID_Configure raid_get_working_config_from_ld_id(PRAID_Core pRaidCore, MV_U8 ldId);
MV_U8 raid_get_data_disk_number(MV_U8 raidLevel, MV_U8 raidDiskNum, MV_U8 parityDiskNum);

void raid_clear_bga(PRAID_Configure pRaidConf);
void raid_set_bga_state(PRAID_Core pRaidCore, MV_U16 ldId, MV_U8 control, MV_U8 bgaType);

MV_U8 raid_is_RDisk_working(PRAID_Disk pRaidDisk);
void raid_set_RDisk_param(PRAID_Disk pRDisk, MV_U8 diskIdx, MV_U8 state, MV_U8 type);

MV_U8 raid_get_ldId_for_block_disk(PRAID_Core pRaidCore, MV_U8 blockIdx);
MV_U8 raid_is_block_disk_wanted(PBlock_Disk pBlkDisk, MV_U16 deviceId);
MV_U8 raid_is_block_disk_available(PBlock_Disk pBlkDisk, MV_U16 deviceId);
MV_U8 raid_block_disk_split(PRAID_Core pRaidCore, PBlock_Disk pSplitBlkDisk, MV_LBA needSize);
MV_U8 raid_disk_fitting(PRAID_Core pRaidCore, MV_LBA needSize, MV_U16 deviceId, MV_PU8 pFitBlkIdx);
MV_U8 raid_block_disk_generate(
	PRAID_Core	pRaidCore,
	MV_LBA		newStartLBA,
	MV_LBA		newSize,
	MV_U16		deviceId
);

/* Won't init this SG table */
void raid_append_SGTable(
	PMV_SG_Table	pSGTable,
	PMV_SG_Entry	*p_pOrgSGEntry,
	MV_PU32			pOffsetInByte,
	MV_U32			sectorCount
	);

/* Will init this SG table */
void raid_generate_SGTable(
	PMV_SG_Table	pNewSGTable,
	PMV_SG_Entry	*p_pOrgSGEntry,
	MV_PU32			pOffsetInByte,
	MV_U32			sectorCount
	);

#ifdef SUPPORT_RAID6
void raid_append_SGTable_from_scratch_buffer(
	PMV_SG_Table	pNewSGTable,
	PScratch_Buffer	pScratchBuffer,
	MV_U32			sectorCount);

void raid_generate_SGTable_from_scratch_buffer(
	PMV_SG_Table	pNewSGTable,
	PScratch_Buffer	pScratchBuffer,
	MV_U32			sectorCount);
#endif

void raid_SGTable_Adjust(PMV_Request pMvReq, MV_U32 sectorCount);

void raid_submit_req_to_next_md(PRAID_Core pRaidCore, PMV_Request pMvReq);
PBlock_Disk raid_get_working_block_disk(
	PRAID_Core		pRaidCore,
	PRAID_Configure	pRaidConf,
	PMap_Info		pMapNode,
	MV_U8			opCode
);

PPD_Disk raid_get_working_pd(PRAID_Core pRaidCore, MV_U8 pdIdx);
MV_U8 raid_is_pd_exist(PRAID_Core pRaidCore, MV_U16 deviceId, MV_PU8 pPDIdx);
MV_U8 raid_is_pd_assigned(PRAID_Core pRaidCore, MV_U16 deviceId);

#ifdef SUPPORT_FREE_POLICY
void raid_set_bypass_device(PRAID_Core pRaidCore, MV_U16 deviceId, MV_U8 pdIndex);
void raid_remove_bypass_device(PRAID_Core pRaidCore, MV_U16 deviceId);
#endif

#ifdef SUPPORT_EVENT
MV_U32 raid_get_bga_event_id(MV_U8 control, MV_U8 bgaType);
#endif

MV_U8 CoreDriver_get_disk_queue_length(MV_U16 deviceID);

#endif /* __RAID_UTIL_H__ */
