
#include "mv_include.h"
#include "mv_os.h"
#include "osd_header.h"
//#include "osd_mod.h"
//#include "osd_inter.h"
#include "linux_main.h"
#include "linux_sense.h"
#include "linux_helper.h"

void GenerateSGTable(
	IN POSD_Extension pOSD,
	IN struct scsi_cmnd *SCpnt,
	OUT PMV_SG_Table pSGTable
	);
void OSDRequestCallback(
	MV_PVOID This,
	PMV_Request pReq
	);

MV_BOOLEAN TranslateSCSIRequest(POSD_Extension pOSD, struct scsi_cmnd *pSCmd, PMV_Request pReq)
{
	
	pReq->Device_Id = mv_scmd_target(pSCmd);	

	/* Cmd_Flag */	//TBD: For Linux: Is that possible to set these flags or need read the Cdb
	pReq->Cmd_Flag = 0;

	/*
	 * Set three flags: CMD_FLAG_NON_DATA, CMD_FLAG_DATA_IN and CMD_FLAG_DMA
	 */
	if ( pSCmd->request_bufflen==0 ) //TBD lily
	{
		pReq->Cmd_Flag |= CMD_FLAG_NON_DATA;
	}
	else
	{
		//if ( Srb->SrbFlags&SRB_FLAGS_DATA_IN )
		//	pReq->Cmd_Flag |= CMD_FLAG_DATA_IN; TBD ?? Lily
		/*We need to optimize the flags setting. Lily*/
		if(SCSI_IS_READ(pSCmd->cmnd[0]))
			pReq->Cmd_Flag |= CMD_FLAG_DATA_IN; /*NOTE!possible to result in ERROR */
		if ( SCSI_IS_READ(pSCmd->cmnd[0]) || SCSI_IS_WRITE(pSCmd->cmnd[0]) )
			pReq->Cmd_Flag |= CMD_FLAG_DMA;
	}

	pReq->Sense_Info_Buffer_Length = SCSI_SENSE_BUFFERSIZE;  //TBD
	pReq->Data_Transfer_Length = pSCmd->request_bufflen;

	//To handle some special CMDs,lily
	memset(pReq->Cdb, 0, MAX_CDB_SIZE);
	switch(pSCmd->cmnd[0]){
	case MODE_SENSE:
		pReq->Cdb[0] = MODE_SENSE_10;
		pReq->Cdb[2] = pSCmd->cmnd[2];
		pReq->Cdb[8] = pSCmd->cmnd[4];
	case REQUEST_SENSE:
		pReq->Cmd_Flag = CMD_FLAG_DATA_IN | CMD_FLAG_DMA;
		break;
		
	case MODE_SELECT:
		pReq->Cdb[0] = MODE_SELECT_10;
		pReq->Cdb[1] = pSCmd->cmnd[1];
		pReq->Cdb[8] = pSCmd->cmnd[4];
		pReq->Cmd_Flag = CMD_FLAG_DATA_IN | CMD_FLAG_DMA;
		break;
		
	case FORMAT_UNIT:
		pReq->Cdb[0] = 0x24; //ATAPI opcodes
		break;
		
	case READ_CAPACITY: //TBD
		pReq->Cdb[0] = pSCmd->cmnd[0];
		pReq->Cmd_Flag = CMD_FLAG_DATA_IN | CMD_FLAG_DMA;
		break;

	case TEST_UNIT_READY:                       //TBD
		pReq->Cdb[0] = pSCmd->cmnd[0];
		break;
		
	default:
		memcpy(pReq->Cdb, pSCmd->cmnd, MAX_CDB_SIZE);
		break;
	}
	
	if (SCSI_IS_INSTANT(pSCmd->cmnd[0]) && pSCmd->use_sg) {
		struct scatterlist *sg = (struct scatterlist *) pSCmd->request_buffer;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 14)
		pReq->Data_Buffer = kmalloc(sg->length, GFP_ATOMIC);
		if (pReq->Data_Buffer) {
			memset(pReq->Data_Buffer, 0, sg->length);
		}
#else
		pReq->Data_Buffer = kzalloc(sg->length, GFP_ATOMIC);
#endif /* 2.6.17 */
		if ( NULL == pReq->Data_Buffer )
			return MV_FALSE;

		pReq->Data_Transfer_Length = sg->length;
		MV_SCp(pSCmd)->map_atomic = 1;
//#elif //TBD
#endif /* 2.5.0  */
	} else {
		pReq->Data_Buffer = pSCmd->request_buffer;
	}

	pReq->Sense_Info_Buffer = pSCmd->sense_buffer;

	/* Init the SG table first no matter it's data command or non-data command. */
	SGTable_Init(&pReq->SG_Table, 0);
	if ( pSCmd->request_bufflen )
	{
		GenerateSGTable(pOSD, pSCmd, &pReq->SG_Table);
	}

	pReq->Org_Req = pSCmd;
	pReq->Context = NULL;

	pReq->LBA.value = 0;
	pReq->Sector_Count = 0;

	pReq->Tag = pSCmd->tag; 
	pReq->Scsi_Status = REQ_STATUS_PENDING;

	pReq->Completion = OSDRequestCallback;

	return MV_TRUE;
}

MV_BOOLEAN TranslateOSRequest(
	IN POSD_Extension pOSD,
	IN struct scsi_cmnd * pSCmd,
	OUT PMV_Request pReq
	)
{
	pReq->Cmd_Initiator = pOSD; //TODO
	pReq->Org_Req = pSCmd;
	pReq->Scsi_Status = REQ_STATUS_INVALID_REQUEST; //TBD??

 	return TranslateSCSIRequest(pOSD, pSCmd, pReq);
			
}

/* This is the only function that OS request can be returned. */
void OSDRequestCallback(
	MV_PVOID This,
	PMV_Request pReq
	)
{
	POSD_Extension pOSD = (POSD_Extension)This;
	struct scsi_cmnd *pSCmd = (struct scsi_cmnd *)pReq->Org_Req;
#ifdef __AC_RAID__
	extern unsigned long __ac_raid_req_count;

	__ac_raid_req_count--;
#endif /* __AC_RAID__ */
	/* Return this request to OS. */
	OSD_Translate_Req_Status_To_OS_Status(pOSD, pSCmd, pReq);
	
	List_Add(&pReq->Queue_Pointer, &pOSD->Free_Request);
	pOSD->Io_Count--;

}

void GenerateSGTable(
	IN POSD_Extension pOSD,
	IN struct scsi_cmnd *SCpnt,
	OUT PMV_SG_Table pSGTable
	)
{
	struct scatterlist *sg = (struct scatterlist *)SCpnt->request_buffer;
 	unsigned int sg_count = 0;
 	BUS_ADDRESS busaddr = 0;
  	int i;

	MV_DBG(DMSG_FREQ,
	       "In GenerateSGTable.\n");
	
	 if (SCpnt->request_bufflen > (mv_scmd_host(SCpnt)->max_sectors << 9))
    	{
       	 MV_PRINT("ERROR: request length exceeds the maximum alowed value, %x %x\n",
              	 pOSD->Device_Id, pOSD->Revision_Id);
    	}

	if (SCpnt->use_sg) {
		 unsigned int length;
		sg = (struct scatterlist *) SCpnt->request_buffer;
		if(MV_SCp(SCpnt)->mapped == 0)
		{
			MV_DBG(DMSG_FREQ,"__MV__ call pci_map_sg.\n");
			sg_count = pci_map_sg(pOSD->pcidev, sg,
	                                SCpnt->use_sg,
	                                scsi_to_pci_dma_dir(SCpnt->sc_data_direction));
			if (sg_count != SCpnt->use_sg)
	            		MV_PRINT("WARNING sg_count(%d) != SCpnt->use_sg(%d)\n",
	                		   (unsigned int)sg_count, SCpnt->use_sg);
			MV_SCp(SCpnt)->mapped = 1;
		}
	
		for (i = 0; i < sg_count; i++) {
			busaddr = sg_dma_address(&sg[i]);
			length = sg_dma_len(&sg[i]);
			/* merge the sg elements if possible */
			/*if (i && last==busaddr && pSGTable->size &&
					pSGTable->size+length<=0x10000 && (addr & 0xffffffff)!=0) {
					//pSGTable->size += length;
					//last += length; //TODO ????
			}//TBD lily
			else*/ {
				if (busaddr & 1)
				{
					MV_ASSERT(0);
					return;//TODO
				}
	
				SGTable_Append( pSGTable, 
						LO_BUSADDR(busaddr), 
						HI_BUSADDR(busaddr),
						length
						);
			}
		}
	}else {
		if(MV_SCp(SCpnt)->mapped == 0)
		{
			MV_DPRINT(("pci_map_single\n"));
			MV_SCp(SCpnt)->bus_address = busaddr = pci_map_single(pOSD->pcidev,
				SCpnt->request_buffer,
				SCpnt->request_bufflen,
				scsi_to_pci_dma_dir(SCpnt->sc_data_direction));

			if (busaddr & 0x1)
			{
				MV_ASSERT(0);
				return; //TODO
			}

			MV_SCp(SCpnt)->mapped = 1;
		}
	
		SGTable_Append( pSGTable, 
						LO_BUSADDR(busaddr), 
						HI_BUSADDR(busaddr),
						SCpnt->request_bufflen);
		
	}
	
}

/*need to be optimized lily*/
void OSD_kunmap_sg(void* pReq)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	void *buf;
	struct scsi_cmnd *scmd =NULL;
	struct scatterlist *sg =NULL;
	PMV_Request        req =NULL;

	req  = (PMV_Request) pReq;
	scmd = (struct scsi_cmnd *) req->Org_Req;
	sg   = (struct scatterlist *) scmd->request_buffer;

	if (MV_SCp(scmd)->map_atomic) {
		WARN_ON(!irqs_disabled());
		buf = kmap_atomic(sg->page, KM_IRQ0) + sg->offset;
		memcpy(buf, req->Data_Buffer, sg->length);
		kunmap_atomic(buf, KM_IRQ0);
		kfree(req->Data_Buffer);
		MV_SCp(scmd)->map_atomic = 0;
	}
#endif	
}

static void osd_shutdown_req_cb(MV_PVOID this, PMV_Request req)
{
	POSD_Extension posd = (POSD_Extension) this;
#ifdef __AC_RAID__
	extern unsigned long __ac_raid_req_count;

	__ac_raid_req_count--;
#endif /* __AC_RAID__ */
	List_Add(&req->Queue_Pointer, &posd->Free_Request);
	posd->Io_Count--;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11)
	atomic_set(&posd->osd_sync, 0);
#else
	complete(&posd->cmpl);
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11) */
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11)
/* will wait for atomic value atomic to become zero until timed out */
/* return how much 'timeout' is left or 0 if already timed out */
int __osd_wait_for_atomic_timeout(atomic_t *atomic, unsigned long timeout)
{
	unsigned intv = HZ/20; 

	while (timeout) {
		if ( 0 == atomic_read(atomic) )
			break;

		if ( timeout < intv )
			intv = timeout;
		set_current_state(TASK_INTERRUPTIBLE);
		timeout -= (intv - schedule_timeout(intv));
	}
	return timeout;
}
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11) */

void osd_send_shutdown_req(POSD_Extension posd)
{
        unsigned long flags;
        PMV_Request pReq;

        /*Send MV_REQUEST to do something.*/	
        pReq = kmalloc(sizeof(MV_Request), GFP_ATOMIC);	
        pReq->Cmd_Initiator = posd;
        pReq->Org_Req = pReq; /*no ideas.*/
        pReq->Scsi_Status = REQ_STATUS_INVALID_REQUEST;
        pReq->Completion = osd_shutdown_req_cb;
	
#ifdef RAID_DRIVER
        pReq->Cdb[0] = APICDB0_LD;
        pReq->Cdb[1] = APICDB1_LD_SHUTDOWN;
#else
        pReq->Device_Id = 0;
        pReq->Cmd_Flag = 0;
        pReq->Cmd_Flag |= CMD_FLAG_NON_DATA;
        pReq->Sense_Info_Buffer_Length = 0;  
        pReq->Data_Transfer_Length = 0;
        pReq->Data_Buffer = NULL;
        pReq->Sense_Info_Buffer = NULL;
        SGTable_Init(&pReq->SG_Table, 0);
        pReq->Cdb[0] = SCSI_CMD_MARVELL_SPECIFIC;
        pReq->Cdb[1] = CDB_CORE_MODULE;
        pReq->Cdb[2] = CDB_CORE_SHUTDOWN;
        pReq->Context = NULL;
        pReq->LBA.value = 0;
        pReq->Sector_Count = 0;
        pReq->Scsi_Status = REQ_STATUS_PENDING;
#endif

        spin_lock_irqsave(posd->lock, flags);
        List_AddTail(&pReq->Queue_Pointer, &posd->Waiting_Request);
        posd->Io_Count++;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11)
	atomic_set(&posd->osd_sync, 1);
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11) */
        OSD_HandleWaitingList(posd);
        spin_unlock_irqrestore(posd->lock, flags);

        spin_lock_irqsave(posd->lock, flags);
        del_timer_sync(&posd->timer);
        spin_unlock_irqrestore(posd->lock, flags);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11)
	__osd_wait_for_atomic_timeout(&posd->osd_sync, 10*HZ);
#else
	wait_for_completion_timeout(&posd->cmpl, 10*HZ);
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11) */
}

MV_BOOLEAN OSD_CanHandleRequest (PMV_Request pReq)
{
	switch ( pReq->Cdb[0] )
	{
		case APICDB0_ADAPTER:
			if ( pReq->Cdb[1] == APICDB1_ADAPTER_GETINFO )
				return MV_TRUE;
			else
				return MV_FALSE;
#ifdef SUPPORT_EVENT
		case APICDB0_EVENT:
			return MV_TRUE;
#endif  /* SUPPORT_EVENT */
		default:
			return MV_FALSE;
	}
}

void OSD_HandleWaitingList(POSD_Extension pOSD)
{
	PMV_Request pReq = NULL;
	MV_PVOID pNextExtension = NULL;
	MV_VOID (*pNextFunction)(MV_PVOID , PMV_Request) = NULL;
#ifdef __AC_RAID__
	unsigned int count =0;
	extern unsigned long __ac_raid_req_count;
#endif /* __AC_RAID__ */

	/* Get the request header */
	while ( !List_Empty(&pOSD->Waiting_Request) ) {
		pReq = (PMV_Request)List_GetFirstEntry(&pOSD->Waiting_Request,
						       MV_Request, 
						       Queue_Pointer);
		MV_DASSERT( pReq != NULL );

		if ( NULL == pReq )
			break;
#if 0
		pCore = pOSD->Module_Manage.resource[MODULE_CORE].module_extension;
		//TBD: To the lower module
		module_set[MODULE_CORE].module_submitrequest(pCore, pReq);
#else
		if ( OSD_CanHandleRequest(pReq) ) {
			OSD_ModuleSubmitRequest( pOSD, pReq );
		} else {
			//TBD: performance
			OSD_GetNextModuleSubmitFunction(pOSD, 
							&pNextExtension, 
							&pNextFunction);
			MV_DASSERT( pNextExtension!=NULL );
			MV_DASSERT( pNextFunction!=NULL );
			pNextFunction(pNextExtension, pReq);
#ifdef __AC_RAID__
			__ac_raid_req_count++;
			count++;
#endif /* __AC_RAID__ */
		}
#endif
	}
#ifdef __AC_RAID__
#define THRESH_HOLD 1
	if  ( count > THRESH_HOLD ) {
		MV_DBG(DMSG_RAID, "__MV__ OSD : sent %d reqs this round.\n", 
		       count);
	}
#endif /* __AC_RAID__ */
}
