/*
 * Linux Driver for 61xx
 * Copyright (C) 2006 Marvell Technology Group Ltd.. All Rights Reserved.
 * linux_main.c
 * lily initialized on Feb. 15 2006
 *
 *  ioctl handler has been implemented.
 *  June 2006, Zax Liu < zaxl at marvell dot com >
 *
 *  implement ioctl the 2.6.11 plus way ( not rely on BKL )
 *  July 2006, Albert Ke < ake at marvell dot com >
 *
 */

#include "mv_os.h"
#include "mv_include.h"

#include "osd_header.h"

#include "linux_main.h"
#include "linux_iface.h"
#include "linux_helper.h"

#include "com_define.h" 
#include "com_type.h"
#include "mv_config.h"
#include "mv_include.h"

MODULE_AUTHOR ("Marvell Technology Group Ltd.");
MODULE_DESCRIPTION ("RAID driver");
MODULE_LICENSE("GPL");

/* 
 * module parameter 
 *
 * refer to ../common/com_dbg.h 
 */
unsigned int mv_dbg_opts = 0;
module_param(mv_dbg_opts, uint, S_IRWXUGO);

#ifdef __BIG_LOCK__
static spinlock_t mv_big_lock;
#endif /* __BIG_LOCK__ */

static const struct pci_device_id __devinitdata mv_pci_ids[] = {
	{PCI_DEVICE(VENDOR_ID, DEVICE_ID_THORLITE_0S1P)},
	{PCI_DEVICE(VENDOR_ID, DEVICE_ID_THORLITE_2S1P)},
	{PCI_DEVICE(VENDOR_ID, DEVICE_ID_THOR_4S1P)},
	{PCI_DEVICE(VENDOR_ID, DEVICE_ID_THOR_4S1P_NEW)},
	{PCI_DEVICE(VENDOR_ID, DEVICE_ID_THORLITE_2S1P_WITH_FLASH)},
	{0}
};

#ifdef __AC_INIT__
MODULE_DEVICE_TABLE(pci, mv_pci_ids);
#endif /* __AC_INIT__ */

POSD_Extension g_Linux_OSD = NULL;
POSD_Extension mv_device_extension_list[MV_DEVICE_MAX_SLOT] = {NULL};
unsigned int mv_device_count = 0;
int mv_major = -1;
extern struct file_operations mv_fops;

#ifdef __AC_RAID__
unsigned long __ac_raid_req_count =0;
#endif /* __AC_RAID__ */

static void timeout_sem(unsigned long data)
{
        up((struct semaphore *)data);
}

static int osd_wait_for_irq_timeout(POSD_Extension pOSD,
				unsigned long time)
{
        struct timer_list timer;
        int f  = 0; 

        if (down_trylock(&pOSD->sem)) {

                init_timer(&timer);
                timer.expires = jiffies + time * HZ;
                timer.data = (unsigned long)&pOSD->sem;
                timer.function = timeout_sem;
                add_timer(&timer);

                if (down_interruptible(&pOSD->sem)) {
                        MV_PRINT("MV : Interrupted.\n");
                        f = 1;
                }
        
                del_timer_sync(&timer);

                if (f) {
                        up(&pOSD->sem);
                        MV_PRINT("MV : Release Semaphore.\n");
                        return 1; /* timed out */
                }
        }
        return 0; /* Interrupted */
}

static void release_host(POSD_Extension pOSD)
{
        unsigned long flags;
        int i;
        struct pci_dev *pcidev = pOSD->pcidev;
        PMV_Request pReq;

        /*Send MV_REQUEST to do something.*/	
        pReq = kmalloc(sizeof(MV_Request), GFP_ATOMIC);	
        pReq->Cmd_Initiator = pOSD;
        pReq->Org_Req = pReq;/*no ideas.*/
        pReq->Scsi_Status = REQ_STATUS_INVALID_REQUEST;
        pReq->Completion = IOOSDRequestCallback;
	
#ifdef RAID_DRIVER
        pReq->Cdb[0] = APICDB0_LD;
        pReq->Cdb[1] = APICDB1_LD_SHUTDOWN;
#else
        pReq->Device_Id = 0;
        pReq->Cmd_Flag = 0;
        pReq->Cmd_Flag |= CMD_FLAG_NON_DATA;
        pReq->Sense_Info_Buffer_Length = 0;  
        pReq->Data_Transfer_Length = 0;
        pReq->Data_Buffer = NULL;
        pReq->Sense_Info_Buffer = NULL;
        SGTable_Init(&pReq->SG_Table, 0);
        pReq->Cdb[0] = SCSI_CMD_MARVELL_SPECIFIC;
        pReq->Cdb[1] = CDB_CORE_MODULE;
        pReq->Cdb[2] = CDB_CORE_SHUTDOWN;
        pReq->Context = NULL;
        pReq->LBA.value = 0;
        pReq->Sector_Count = 0;
        pReq->Scsi_Status = REQ_STATUS_PENDING;
#endif

        spin_lock_irqsave(pOSD->lock, flags);
        List_AddTail(&pReq->Queue_Pointer, &pOSD->Waiting_Request);
        pOSD->Io_Count++;
        OSD_HandleWaitingList(pOSD);
        spin_unlock_irqrestore(pOSD->lock, flags);

        spin_lock_irqsave(pOSD->lock, flags);
        del_timer_sync(&pOSD->timer);
        spin_unlock_irqrestore(pOSD->lock, flags);

        osd_wait_for_irq_timeout(pOSD, 1);
        
        free_irq(pOSD->pcidev->irq, pOSD);

        for ( i=0; i<MAX_BASE_ADDRESS; i++ )
                if (pci_resource_flags(pcidev, i) & IORESOURCE_MEM)
                        iounmap(pOSD->Base_Address[i]);

        pci_release_regions(pcidev);
        pci_disable_device(pcidev);

        for ( i=0; i<MAX_MODULE_NUMBER; i++ ) {
                if ( pOSD->uncached_virtual_address[i] ) {
                        kfree(pOSD->uncached_virtual_address[i]);
                }
        }
			
        vfree(pOSD->host_data); /*pOSD is invalid*/
        pOSD = NULL;
        mv_device_count--;
        kfree(pReq);
}


/* notifier block to get notified on system shutdown/halt/reboot */
static int mv_linux_halt(struct notifier_block *nb, unsigned long event,
			 void *buf)
{
	POSD_Extension pOSD = NULL;

	switch (event) {
	case SYS_RESTART:
	case SYS_HALT:
	case SYS_POWER_OFF:
		while (g_Linux_OSD) {
			pOSD = g_Linux_OSD;
			g_Linux_OSD = g_Linux_OSD->next;
			osd_send_shutdown_req(pOSD);
		}
		break;
	default:
		break;
	}
	return NOTIFY_OK;
}

static struct notifier_block mv_linux_notifier = {
	mv_linux_halt, NULL, 0
};


static irqreturn_t mv_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	/* MV_FALSE should be equal to IRQ_NONE (0) */
	irqreturn_t retval = MV_FALSE;
	unsigned long flags;
	MV_PVOID pcore;
 
	POSD_Extension pOSD = (POSD_Extension)dev_id;
	PModule_Manage module_manage = &pOSD->Module_Manage;
	

	spin_lock_irqsave(pOSD->lock, flags);
	pcore = module_manage->resource[MODULE_CORE].module_extension;
	retval = Core_InterruptServiceRoutine(pcore);
	spin_unlock_irqrestore(pOSD->lock, flags);

	MV_DBG(DMSG_FREQ, "__MV__ Exit intr handler.\n");

	return IRQ_RETVAL(retval);
}

extern struct scsi_host_template driver_template;
static int mv_probe_device(struct pci_dev *pcidev,
				      const struct pci_device_id *id)
{

	MV_U8			i;
	MV_U32			size;
	MV_U16			requestNumber = MAX_REQUEST_NUMBER;
	struct Scsi_Host	*phost=NULL;
	PModule_Header		pheader = NULL;
	POSD_Extension		pOSD = NULL;
	PModule_Manage		pmodule_manage = NULL;
	MV_U32			uncached_requirement = 0;
	MV_PVOID		uncached_virtual = NULL;
	MV_PHYSICAL_ADDR	uncached_physical;
#ifndef __BIG_LOCK__
	spinlock_t		driver_lock;
#endif /* __BIG_LOCK__ */
	
	pci_set_drvdata(pcidev, NULL);

	if (pci_enable_device(pcidev)) {
		MV_PRINT("pci_enable_device() failed\n");
		return -ENODEV;
	}

	pci_set_master(pcidev);


    	if (pci_set_dma_mask(pcidev, 0xffffffffffffffffULL)) {
		if (pci_set_dma_mask(pcidev, 0xffffffffUL)) {
			MV_DBG(DMSG_KERN, 
			       "__MV__ Cannot set DMA mask.\n");
        			pci_disable_device(pcidev);
			return -ENODEV;
		} else { 
			MV_DBG(DMSG_KERN, "__MV__ Using 32-bit DMA.\n");
		}
     	} else { 
		MV_DBG(DMSG_KERN,"__MV__ Using 64-bit DMA.\n");
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	if (pci_request_regions(pcidev, mv_driver_name) != 0) {//TODO?
		MV_DBG(DMSG_KERN, "__MV__ pci_request_regions() failed.\n");
		pci_disable_device(pcidev);
		return -ENOMEM;
    	}
#endif

	size = mod_get_mem_size(pOSD, RESOURCE_CACHED_MEMORY, (MV_U16)requestNumber);

	pheader = (PModule_Header) vmalloc(size); //test lily

	MV_ASSERT(pheader!=NULL);

	if (pheader == NULL) {
		MV_PRINT("Adapter allocation failed\n");
 failed:
 		if (pheader) vfree(pheader);
 		for ( i=0; i<MAX_MODULE_NUMBER; i++ )
			if(pOSD->uncached_virtual_address[i]) kfree(pOSD->uncached_virtual_address[i]);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		pci_release_regions(pcidev);
#endif
		pci_disable_device(pcidev);
		//free() TO DO
		return -ENOMEM;
    	}
    	
    	memset(pheader, 0, size);
    	
	Module_AssignModuleExtension(pheader, requestNumber);
	pOSD = (POSD_Extension)head_to_osd(pheader);
	pmodule_manage = &pOSD->Module_Manage;
	pOSD->host_data = pheader;
	pOSD->pcidev = pcidev;

	pOSD->Vendor_Id = pcidev->vendor;
	pOSD->Device_Id = pcidev->device;

	if ( pci_read_config_byte(pcidev, PCI_REVISION_ID, &pOSD->Revision_Id) ) {
		MV_PRINT("mv61xx: Failed to get revision id.\n");
		goto failed;
    	}
	
	for ( i=0; i<MAX_BASE_ADDRESS; i++ ) {
		unsigned long base = pci_resource_start(pcidev, i);
		unsigned long length = pci_resource_len(pcidev, i);
		
		if (pci_resource_flags(pcidev, i) & IORESOURCE_MEM)
			pOSD->Base_Address[i] = (MV_PVOID)ioremap(base, length);
		else 
			pOSD->Base_Address[i] =  (MV_PVOID)base;

		MV_DPRINT(("Base Address %d = %x\n", i, (unsigned int)  (MV_PTR_INTEGER)pOSD->Base_Address[i] ));
	}
	
	uncached_requirement = mod_get_mem_size(pOSD,
						RESOURCE_UNCACHED_MEMORY, 
						requestNumber);
	//uncached_virtual = kmalloc(uncached_requirement, GFP_ATOMIC);
	
	//MV_ASSERT(uncached_virtual!=NULL);

	if (uncached_requirement == -1) {
		MV_PRINT("Uncached memory allocation failed\n");
		goto failed;
	}

	//pOSD->uncached_virtual_address = uncached_virtual;
	for ( i=0; i<MAX_MODULE_NUMBER; i++ ) {
		BUS_ADDRESS tmpaddr;
		
		uncached_virtual = pOSD->uncached_virtual_address[i];
		tmpaddr = (BUS_ADDRESS) virt_to_bus(uncached_virtual);
		
		uncached_physical.low = LO_BUSADDR(tmpaddr);
		uncached_physical.high = HI_BUSADDR(tmpaddr);
	
		Module_AssignUncachedMemory( pmodule_manage, 
					     uncached_virtual, 
					     uncached_physical, 
					     uncached_requirement, 
					     requestNumber, 
					     i);
	}


#ifdef __BIG_LOCK__
	pOSD->lock = &mv_big_lock;
#else
	spin_lock_init(&driver_lock);
	pOSD->lock = &driver_lock;
#endif /* __BIG_LOCK__ */

	Module_InitializeAll(pmodule_manage, requestNumber);

	init_timer(&pOSD->timer);

	sema_init(&pOSD->sem, 0);
	
	init_completion(&pOSD->cmpl);
	
	spin_lock_irq(pOSD->lock);
	Module_StartAll(pmodule_manage, MODULE_CORE);
	spin_unlock_irq(pOSD->lock);				

	phost = scsi_register(&driver_template, sizeof(void *));

	if (!phost) {
		MV_PRINT("scsi_register failed");
		goto failed;
	}

	if (g_Linux_OSD  == NULL) {
		g_Linux_OSD = pOSD;
	} else {
		pOSD->next = g_Linux_OSD;
		g_Linux_OSD = pOSD;
	}
	
	get_osd_ext_header(phost) = pheader;
	pOSD->host = phost;
    	
    	phost->irq = pcidev->irq;

	phost->max_id = MV_MAX_TARGET_NUMBER;
	phost->max_lun = MV_MAX_LUN_NUMBER;
	phost->max_channel = 0;
	phost->max_cmd_len = 16;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	pOSD->lock = &io_request_lock;
#else 
#ifndef __BIG_LOCK__
	pOSD->lock = phost->host_lock;
#endif /* __BIG_LOCK__ */
#endif

	if (request_irq(pcidev->irq, mv_intr_handler, SA_SHIRQ, mv_driver_name, pOSD)<0) {
		MV_PRINT("Error requesting IRQ %d", pcidev->irq);
		scsi_unregister(phost);
		goto failed;
	
	}

	if (osd_wait_for_irq_timeout(pOSD, 300)) {
		free_irq(pcidev->irq, pOSD);
		goto failed;
	}

	if (mv_device_count == 0) {
		register_reboot_notifier(&mv_linux_notifier); 
	}

	if (mv_major == -1) {
		if ((mv_major = register_chrdev(0, mv_driver_name, &mv_fops)) < 0)
			MV_PRINT("Failed to register character device");
	}

	mv_device_extension_list[mv_device_count] = pOSD;
	mv_device_count++;

	return 0;
}

static int mv_linux_detect (struct scsi_host_template *pSHTpt)
{
	int                    num_hosts=0;
	int                    index;
	struct pci_dev         *pcidev = NULL;
	const struct pci_device_id   *id = &mv_pci_ids[0];

	index = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_unlock_irq (&io_request_lock);
#endif

	while (1) {
		if ( id[index].device == 0) {
			break;
		}
		pcidev = NULL;

		while ((pcidev = pci_find_device (VENDOR_ID,
	                                  id[index].device, pcidev)) != NULL) {

			MV_PRINT("PCI device found, DeviceId 0x%x BAR0=%lx\n",
				 (unsigned int) id[index].device, 
				 (unsigned long) pci_resource_start(pcidev,0));
			if (mv_probe_device(pcidev, &id[index]) == 0)
				num_hosts ++;
		}
		index ++;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	spin_lock_irq (&io_request_lock);
#endif
	return num_hosts;

}

#ifdef __AC_INIT__
static POSD_Extension init_osd_ext(void)
{
	int i = 0;
	POSD_Extension  posd;
	unsigned int    size;
	PModule_Header  pheader = NULL;


	size = mod_get_mem_size(posd, RESOURCE_CACHED_MEMORY, 
				MAX_REQUEST_NUMBER);
	
	pheader = (PModule_Header) vmalloc(size);
	if ( NULL == pheader ) {
		return pheader;
	}
		
	return NULL;
}

static int mv_probe(struct pci_dev *dev, const struct pci_device_id *id)
{
	unsigned int ret = PCIBIOS_SUCCESSFUL;
	struct Scsi_Host *psht = NULL;
	POSD_Extension   posd  = NULL;
	
	
	ret = pci_enable_device(dev);
	if (ret) {
		MV_DBG(DMSG_KERN, "__MV__ Cannot enable the device.\n");
		return ret;
	}
	
	ret = pci_request_regions(dev, MV_DRV_NAME);
	if (ret)
		goto err_req_region;
	
	
	if (pci_set_dma_mask(dev, 0xFFFFFFFFFFFFFFFFULL)) {
		ret = pci_set_dma_mask(dev, 0xFFFFFFFFUL);
		if (ret)
			goto err_dma_mask;
		MV_DBG(DMSG_KERN, "__MV__ 32bit dma mask.\n");
	}
		
	pci_set_master(dev);

	MV_DBG(DMSG_MSG, "Marvell (S)ATA Controller is found, using IRQ %d.\n",
	       dev->irq);
	
	posd = init_osd_ext();
	if ( NULL == posd ) {
		ret = -ENOMEM;
		goto err_dma_mask;
	}
	
	
	return 0;
	
err_dma_mask:
	pci_release_regions(dev);
err_req_region:
	pci_disable_device(dev);

	return ret;
}

static void mv_remove(struct pci_dev *dev)
{
	return;
}

static struct pci_driver mv_pci_driver = {
	.name     = "mv_raid",
	.id_table = mv_pci_ids,
	.probe    = mv_probe,
	.remove   = mv_remove,
};
#endif /* __AC_INIT__ */

static int mv_linux_queue_command(struct scsi_cmnd *pSCmd, 
				  void (*done) (struct scsi_cmnd *))
{
	struct Scsi_Host *phost = mv_scmd_host(pSCmd);
	PModule_Header pheader = get_osd_ext_header(phost);
	POSD_Extension pOSD = (POSD_Extension)head_to_osd(pheader);
	PMV_Request pReq;
	unsigned long flags;

	if ( done == NULL ) {
		MV_PRINT( ": in queuecommand, done function can't be NULL\n");
		return 0;
    	}

	spin_lock_irqsave(pOSD->lock, flags);
	MV_DBG( DMSG_SCSI_FREQ,
		"mv_linux_queue_command %p (%d/%d/%d/%d cdb=(%x-%x-%x))\n", 
		pSCmd, phost->host_no, mv_scmd_channel(pSCmd), 
		mv_scmd_target(pSCmd), mv_scmd_lun(pSCmd),
		*(pSCmd->cmnd), *(pSCmd->cmnd+1), 
		*(pSCmd->cmnd+2) );

	pSCmd->result = 0;
 	pSCmd->scsi_done = done;
	MV_SCp(pSCmd)->bus_address = 0;
	MV_SCp(pSCmd)->mapped = 0;
	MV_SCp(pSCmd)->map_atomic = 0;
	
	if ( mv_scmd_channel(pSCmd) ) {
		pSCmd->result = DID_BAD_TARGET << 16;
		goto done;
	}

	/* 
	 * Get mv_request resource and translate the scsi_cmnd request to mv_request.
	 */
	MV_DASSERT( !List_Empty(&pOSD->Free_Request) );
	pReq = List_GetFirstEntry((&pOSD->Free_Request), MV_Request, Queue_Pointer);
	if ( pReq == NULL ) {
		spin_unlock_irqrestore(pOSD->lock, flags);
		return SCSI_MLQUEUE_HOST_BUSY;
	}
	
	if ( !TranslateOSRequest(pOSD,pSCmd, pReq) ) {
		/* 
		 * Even TranslateOSRequest failed, 
		 * it still should set some of the variables to the MV_Request
		 * especially MV_Request.Org_Req and MV_Request.Scsi_Status;
		 */
		MV_DBG( DMSG_OSD,
			"ERROR - Translation from OS Request failed." );
		/* this is tihs */
		pOSD->Io_Count++;
#ifdef __AC_RAID__
		__ac_raid_req_count++;
#endif /* __AC_RAID__ */
		OSDRequestCallback(pOSD, pReq);
		spin_unlock_irqrestore(pOSD->lock, flags);
		return 0;
	}

	/* 
	 * Queue this request. 
	 * Cannot return with BUSY when core driver is not ready. It'll fail hibernation. 
	 */
	List_AddTail(&pReq->Queue_Pointer, &pOSD->Waiting_Request);
	pOSD->Io_Count++;

	if ( pOSD->State != DRIVER_STATUS_STARTED ) {
		MV_ASSERT(0);
		/*if ( pOSD->State==DRIVER_STATUS_IDLE )
		  {
		  pOSD->State = DRIVER_STATUS_STARTING;
		  Module_StartAll(module_manage, MODULE_CORE);
		  }*/
	}
#ifdef USE_MODULE_CO_RESET
	else if ( module_set[MODULE_OSD].is_hardware_reseting(pOSD) && module_set[MODULE_OSD].waiting_for_idle(pOSD) ) {
		module_set[MODULE_OSD].reset(pOSD);
	}
#endif
	else {
		OSD_HandleWaitingList(pOSD);
	}
	spin_unlock_irqrestore(pOSD->lock, flags);
	return 0;
done:
        pSCmd->scsi_done(pSCmd);
        spin_unlock_irqrestore(pOSD->lock, flags);
        return 0;
}

#if defined(__AC_EH__) && 0
static void ac_dump_info(struct scsi_cmnd *cmd)
{
#ifdef __AC_RAID__
	struct Scsi_Host *host;
	POSD_Extension posd;

	host = cmd->device->host;
	posd = (POSD_Extension) head_to_osd(get_osd_ext_header(host));
	
	MV_DBG(DMSG_OSD, "__MV__ OSD IO Count : %d.\n", posd->Io_Count);
	MV_DBG(DMSG_OSD, "__MV__ OSD submitted req count : %ld.\n", 
	       __ac_raid_req_count);
#endif /* __AC_RAID__ */
}

static int mv_linux_abort(struct scsi_cmnd *cmd)
{
	
	struct Scsi_Host *host;
	POSD_Extension posd;
	int  ret = FAILED;
	
	MV_PRINT("__MV__ abort command %p.\n", cmd);

	return ret;
}


static int mv_linux_reset (struct scsi_cmnd *cmd)
{
	MV_PRINT("__MV__ reset handler %p.\n", cmd);
	return FAILED;
}
#endif /* __AC_EH__ */

static int mv_linux_release (struct Scsi_Host *phost)
{
	POSD_Extension pOSD = (POSD_Extension)head_to_osd(get_osd_ext_header(phost));

	if ( g_Linux_OSD == pOSD ) {
		g_Linux_OSD = pOSD->next;
	} else {
		POSD_Extension ptmp = g_Linux_OSD;
		while ( pOSD != ptmp->next ) {
			ptmp=ptmp->next;
		}
		ptmp->next = pOSD->next;
	}
	
	release_host(pOSD);
	
	if ( g_Linux_OSD == NULL ) {
		unregister_reboot_notifier(&mv_linux_notifier);
	}

	scsi_unregister(phost);
	return 0;
}

static enum scsi_eh_timer_return mv_linux_timed_out(struct scsi_cmnd *cmd)
{
	static int i;

	MV_DBG(DMSG_SCSI, "__MV__ scmd timed out : ");
	MV_DBG( DMSG_SCSI,
		"%p (%d/%d/%d cdb=(%x-%x-%x)).\n", 
		cmd, mv_scmd_channel(cmd), 
		mv_scmd_target(cmd), mv_scmd_lun(cmd),
		*(cmd->cmnd), *(cmd->cmnd+1), 
		*(cmd->cmnd+2) );

	if ( i++ > 5 )
		return EH_NOT_HANDLED;
	else
		return EH_RESET_TIMER;
}

struct scsi_host_template driver_template = {
        module:                         THIS_MODULE,
        name:                           "Marvell 61xx Storage Controller",
        detect:                         mv_linux_detect,
        release:                        mv_linux_release,
        proc_name:                      mv_driver_name,
        proc_info:                      mv_linux_proc_info,
        queuecommand:                   mv_linux_queue_command,
#if defined(__AC_EH__) && 0
        eh_abort_handler:		mv_linux_abort,
        eh_device_reset_handler:	mv_linux_reset,
        eh_bus_reset_handler:		mv_linux_reset,
        eh_host_reset_handler:		mv_linux_reset,
#endif /* __AC_EH__ */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        eh_timed_out:                   mv_linux_timed_out,
#endif
        //bios_param:                   mv_linux_bios_param,
        /* save 4 for ioctl */
        can_queue:                      MAX_REQUEST_NUMBER-2,
        this_id:                        -1,
        max_sectors:                    MV_MAX_TRANSFER_SECTOR,
        sg_tablesize:                   MAX_SG_ENTRY,
        cmd_per_lun:                    MAX_REQUEST_NUMBER-2,
        use_clustering:                 DISABLE_CLUSTERING,
        emulated:                       0
};

static int __init mv_linux_driver_init(void)
{
	struct scsi_host_template *sht = &driver_template;
	struct Scsi_Host *shost;
	struct list_head *l;
	int error;

	if (!sht->release) {
		printk(KERN_ERR
			"scsi HBA driver %s didn't set a release method.\n",
			sht->name);
		return -EINVAL;
	}

        /* default to only show no msg */
#ifdef __AC_DBG__
	mv_dbg_opts |= ( DMSG_CORE|DMSG_OSD|DMSG_KERN|DMSG_ACDB|DMSG_SCSI );
	mv_dbg_opts |= ( DMSG_ERRH|DMSG_RAID );

#ifdef __SCSI_FREQ__
	mv_dbg_opts |= DMSG_SCSI_FREQ;
#endif /* __SCSI_FREQ__ */
#ifdef __AC_ALL_MSG__
        /* even i dont want those FREQs */
	mv_dbg_opts |= DMSG_FREQ | DMSG_SCSI_FREQ;
#endif /* __AC_ALL_MSG__ */

#endif /* __AC_DBG__ */

#ifdef __BIG_LOCK__
        spin_lock_init(&mv_big_lock);
#endif /* __BIG_LOCK__ */

	/* bg thread init - refer to osd_timer.[ch] */
	osd_house_keeper_init();

	sht->module = THIS_MODULE;
	INIT_LIST_HEAD(&sht->legacy_hosts);
	
	if ( sht->detect(sht) == 0 ) {
		error = -ENODEV;
		goto err_osd_hp;
	}

	if (list_empty(&sht->legacy_hosts)) {
		error = -ENODEV;
		goto err_osd_hp;
	}
	

	list_for_each_entry(shost, &sht->legacy_hosts, sht_legacy_list) {
		POSD_Extension pOSD = (POSD_Extension)head_to_osd(get_osd_ext_header(shost));
		MV_DPRINT(("mv_linux_driver_init: host %p\n",shost));
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)	&& defined(__AC_EH__)
                shost->transportt->eh_timed_out = mv_linux_timed_out;
#endif
		error = scsi_add_host(shost, &pOSD->pcidev->dev);
		if (error)
			goto fail;
		scsi_scan_host(shost);
	}
	
	osd_house_keeper_run();

	return 0;
fail:
	l = &shost->sht_legacy_list;
	while ((l = l->prev) != &sht->legacy_hosts)
		scsi_remove_host(list_entry(l, struct Scsi_Host, sht_legacy_list));
err_osd_hp:
	osd_house_keeper_exit();
	return error;
}

static void __exit mv_linux_driver_exit(void)
{
	struct scsi_host_template *sht = &driver_template;
	struct Scsi_Host *shost, *s;
	
	list_for_each_entry(shost, &sht->legacy_hosts, sht_legacy_list) {
		MV_DBG(DMSG_KERN, "mv_linux_driver_exit: host %p.\n", shost);
		scsi_remove_host(shost);
	}

	list_for_each_entry_safe(shost,s,&sht->legacy_hosts,sht_legacy_list) {
		sht->release(shost);
	}

	if (mv_major >= 0) {
		unregister_chrdev(mv_major, mv_driver_name);
	}
	
	osd_house_keeper_exit();

	if (list_empty(&sht->legacy_hosts))
		return;

	printk(KERN_WARNING "%s did not call scsi_unregister\n", sht->name);

	list_for_each_entry_safe(shost, s, &sht->legacy_hosts, sht_legacy_list)
		scsi_unregister(shost);
}

module_init(mv_linux_driver_init);
module_exit(mv_linux_driver_exit);



