#include "mv_include.h"
#include "mv_os.h"

#include "osd_header.h"
#include "linux_main.h"
#include "linux_sense.h"

void OSD_Translate_Req_Status_To_OS_Status(
	IN POSD_Extension pOSD,
	IN struct scsi_cmnd *pSCmd,
	IN PMV_Request pReq
	)
{
    	PSENSE_DATA  senseBuffer = (PSENSE_DATA)pSCmd->sense_buffer;
	
	OSD_kunmap_sg(pReq);

	if(MV_SCp(pSCmd)->mapped)
	{
		if (pSCmd->use_sg)
		{
			MV_DBG(DMSG_FREQ, "__MV__ call pci_unmap_sg.\n");
			pci_unmap_sg(
					pOSD->pcidev,
					(struct scatterlist *)pSCmd->request_buffer,
					pSCmd->use_sg,
					scsi_to_pci_dma_dir(pSCmd->sc_data_direction)
				);
		}
		else
		{
			MV_DBG(DMSG_FREQ,"__MV__ call pci_unmap_single.\n");
			pci_unmap_single(
					pOSD->pcidev,
					MV_SCp(pSCmd)->bus_address,
					pSCmd->request_bufflen,
					scsi_to_pci_dma_dir(pSCmd->sc_data_direction)
				);
		}
	}


		
	MV_DBG(DMSG_SCSI_FREQ,
	       "OSD_Translate_Req_Status_To_OS_Status:"
	       " pReq->Scsi_Status = %x pcmd = %p.\n", 
	       pReq->Scsi_Status, pSCmd);
	
	switch(pReq->Scsi_Status) {
	case REQ_STATUS_SUCCESS:
		pSCmd->result = (DID_OK<<16);
		break;
	case REQ_STATUS_MEDIA_ERROR: //TBD
		pSCmd->result = (DID_BAD_TARGET<<16);
		break;
	case REQ_STATUS_BUSY:
		pSCmd->result = (DID_BUS_BUSY<<16);
		break;
	case REQ_STATUS_NO_DEVICE:
		pSCmd->result = (DID_NO_CONNECT<<16);
		break;
	case REQ_STATUS_HAS_SENSE:
		/* Sense buffer data is valid already. */
		pSCmd->result  = (DRIVER_SENSE << 24) | (DID_OK << 16); //| (CHECK_CONDITION << 1) ;
		senseBuffer->Valid = 1;	
	#ifdef MV_DEBUG
				MV_DPRINT(("Sense: response %x SK %s length %x ASC %x ASCQ %x.\n", 
					((MV_PU8)senseBuffer)[0],
					MV_DumpSenseKey(((MV_PU8)senseBuffer)[2]),
					((MV_PU8)senseBuffer)[7],
					((MV_PU8)senseBuffer)[12],
					((MV_PU8)senseBuffer)[13]
					));
	#endif
		break;
	default:
		pSCmd->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
		break;
	}

	pSCmd->scsi_done(pSCmd);
}

