#include "mv_include.h"
#include "mv_os.h"

#include "osd_header.h"

/* For debug purpose only. */
POSD_Extension gOSD = NULL;

extern Module_Interface module_set[];

/*
 * 
 * Module interface function table
 *
 */
MV_U32 OSD_ModuleGetResourceRequirement(enum Resource_Type type, MV_U16 maxIo)
{
	MV_U32 size = 0;

	/* OSD Extension requirement */
	if (type == RESOURCE_CACHED_MEMORY) {
		/* Fixed memory */
		size = OFFSET_OF(OSD_Extension, Memory_Pool);
		size = ROUNDING(size, 8);

		/* MV_Request pool */
		/* MV_Request is 64bit aligned. */
		size += maxIo * MV_REQUEST_SIZE;

		/* Timer pool */
		size += Timer_GetResourceRequirement(maxIo);

		MV_ASSERT(size == ROUNDING(size, 8));

#ifdef SUPPORT_EVENT
		size += sizeof(Driver_Event_Entry) * MAX_EVENTS;
#endif
		MV_ASSERT(size == ROUNDING(size, 8));

		return size;
	}

	/* OSD doesn't need other kind of memory resource. */
	return 0;
}

void OSD_ModuleInitialize(MV_PVOID This, MV_U32 extension_size, MV_U16 max_io)
{
	POSD_Extension pOSD = (POSD_Extension)This;
	MV_PTR_INTEGER temp = (MV_PTR_INTEGER)pOSD->Memory_Pool;
	MV_U8 i;
	PMV_Request pReq = NULL;

#ifdef SUPPORT_EVENT
	PDriver_Event_Entry pEvent = NULL;
#endif
    
	gOSD = pOSD;

	MV_ASSERT( sizeof(MV_Request)==ROUNDING(sizeof(MV_Request),8) );
	/* 
	 * Initialize data structure however following variables have been set already.
	 *	Device_Extension
	 *	Is_Dump
	 *	Base_Address
	 *	Adapter_Bus_Number and Adapter_Device_Number
	 *	Vendor_Id, Device_Id and Revision_Id
	 */
	pOSD->State = DRIVER_STATUS_IDLE;
	pOSD->Io_Count = 0;
	pOSD->Max_Io = max_io;

	pOSD->Module_Manage.status = 0;

	/* Initialize the free request queue. */
	MV_LIST_HEAD_INIT(&pOSD->Free_Request);
	MV_LIST_HEAD_INIT(&pOSD->Waiting_Request);
	temp = ROUNDING( ((MV_PTR_INTEGER)temp), 8 );
	for ( i=0; i<max_io; i++ )
	{
		pReq = (PMV_Request)temp;
		List_AddTail(&pReq->Queue_Pointer, &pOSD->Free_Request);
		temp += MV_REQUEST_SIZE;	/* MV_Request is 64bit aligned. */
	}	

#ifdef SUPPORT_EVENT
	
	MV_LIST_HEAD_INIT(&pOSD->Stored_Events);
	MV_LIST_HEAD_INIT(&pOSD->Free_Events);
	pOSD->Num_Stored_Events = 0;
	pOSD->SequenceNumber = 0;	// Event sequence number

	MV_ASSERT( sizeof(Driver_Event_Entry)==ROUNDING(sizeof(Driver_Event_Entry),8) );
	temp = ROUNDING( ((MV_PTR_INTEGER)temp), 8 );

	for ( i=0; i<MAX_EVENTS; i++ )
	{
		pEvent = (PDriver_Event_Entry)temp;
		List_AddTail( &pEvent->Queue_Pointer, &pOSD->Free_Events );
		temp += sizeof( Driver_Event_Entry );
	}

#endif

	/* Initialize timer module. */
	Timer_Initialize(&pOSD->Timer_Module, (MV_PU8)temp);

#ifdef SUPPORT_TIMER
	/* kick off the timer */
	Timer_CheckRequest(pOSD->Device_Extension);
#endif
}

void OSD_ModuleStart(MV_PVOID This)
{
	/* There is nothing we need do here. Just finish this function. */
	OSD_ModuleStarted(This);
}

void OSD_ModuleStop(MV_PVOID This)
{
	POSD_Extension pOSD = (POSD_Extension)This;
	/* Clear the OSD data structure. Reset some variables if necessary. */

	/* At this momment, the outstanding request count should be zero. */
	MV_ASSERT(pOSD->Io_Count == 0);
	
	/* Stop the Timer */
	Timer_Stop(&pOSD->Timer_Module);
}

void OSD_ModuleNotification(MV_PVOID This, 
			    enum Module_Event event, 
			    MV_U32 event_param)
{
#ifdef SUPPORT_HOT_PLUG
	osd_msg_insert(This, event, event_param);
#endif /* SUPPORT_HOT_PLUG */
}

void OSD_ModuleSubmitRequest(MV_PVOID This, PMV_Request pReq)
{
	POSD_Extension pOSD = (POSD_Extension)This;
	switch ( pReq->Cdb[0] )
	{
		case APICDB0_ADAPTER:
			if (pReq->Cdb[1] == APICDB1_ADAPTER_GETINFO)
				mvGetAdapterInfo( pOSD, pReq );
			else
				pReq->Scsi_Status = REQ_STATUS_INVALID_REQUEST;
			break;

#ifdef SUPPORT_EVENT
		case APICDB0_EVENT:
			if (pReq->Cdb[1] == APICDB1_EVENT_GETEVENT)
				mvGetEvent( pOSD, pReq );
			else
				pReq->Scsi_Status = REQ_STATUS_INVALID_REQUEST;
			break;
#endif  /* SUPPORT_EVENT */

		default:
			pReq->Scsi_Status = REQ_STATUS_INVALID_REQUEST;
	}
	pReq->Completion(pReq->Cmd_Initiator, pReq);
}

/* helper functions related to OSD_ModuleSubmitRequest */
void mvGetAdapterInfo( MV_PVOID This, PMV_Request pReq )
{
	POSD_Extension pOSD = (POSD_Extension)This;
	PAdapter_Info pAdInfo;

	/* initialize */
	pAdInfo = (PAdapter_Info)pReq->Data_Buffer;
	MV_ZeroMemory(pAdInfo, sizeof(Adapter_Info));

	/* TBD: some info are missing, will fill in later */

//	MV_CopyMemory(pAdInfo->BaseAddress, pOSD->Base_Address, MAX_BASE_ADDRESS * sizeof(MV_PVOID) );

	pAdInfo->DriverVersion.VerMajor = VER_MAJOR;
	pAdInfo->DriverVersion.VerMinor = VER_MINOR;
	pAdInfo->DriverVersion.VerOEM = VER_OEM;
	pAdInfo->DriverVersion.VerBuild = VER_BUILD;

	pAdInfo->SystemIOBusNumber = pOSD->Adapter_Bus_Number;
	pAdInfo->SlotNumber = pOSD->Adapter_Device_Number;
	pAdInfo->VenDevID = pOSD->Vendor_Id;
	pAdInfo->SubVenDevID = pOSD->Device_Id;

	if ( pOSD->Device_Id == DEVICE_ID_THORLITE_2S1P ||
	     pOSD->Device_Id == DEVICE_ID_THORLITE_2S1P_WITH_FLASH )
		pAdInfo->PortCount = 3;
	else if ( pOSD->Device_Id == DEVICE_ID_THORLITE_0S1P )
		pAdInfo->PortCount = 1;
	else
		pAdInfo->PortCount = 5;

	pAdInfo->AlarmSupport = MV_FALSE;
	pAdInfo->MaxBlockPerPD = 8;		/* hardcoded to 8 for now */

	pReq->Scsi_Status = REQ_STATUS_SUCCESS;
}

#ifdef SUPPORT_EVENT

void mvGetEvent( MV_PVOID This, PMV_Request pReq )
{
	POSD_Extension pOSD = (POSD_Extension)This;
	PEventRequest pEventReq = (PEventRequest)pReq->Data_Buffer;
	PDriver_Event_Entry pfirst_event;
	MV_U8 count = 0;

	pEventReq->Count = 0;
	
	if ( pOSD->Num_Stored_Events > 0 )
	{	
		MV_DASSERT( !List_Empty(&pOSD->Stored_Events) );
		while (!List_Empty(&pOSD->Stored_Events) && ( count < MAX_EVENTS_RETURNED))
		{
			pfirst_event = List_GetFirstEntry((&pOSD->Stored_Events), Driver_Event_Entry, Queue_Pointer);
			MV_CopyMemory( &pEventReq->Events[count], &pfirst_event->Event, sizeof (DriverEvent));
			pOSD->Num_Stored_Events--;
			List_AddTail( &pfirst_event->Queue_Pointer, &pOSD->Free_Events );
			count++;
		}
		pEventReq->Count = count;
	}

	pReq->Scsi_Status = REQ_STATUS_SUCCESS;
	return;
}

#endif

void OSD_ModuleReset(MV_PVOID extension)
{
	OSD_ModuleInitialize(extension, sizeof(OSD_Extension), 32);//TBD
}

#ifdef USE_MODULE_CO_RESET
MV_BOOLEAN OSD_WaitingForIdle(MV_PVOID extension)
{
	POSD_Extension pOSD = (POSD_Extension)extension;

	PModule_Manage module_manage = &pOSD->Module_Manage;
	
	MV_U16 moduleId = Module_GetModuleId(extension);


    	while(module_set[++moduleId].waiting_for_idle)
    	{
		MV_ASSERT(module_manage->resource[moduleId].module_extension);
    		return module_set[moduleId].waiting_for_idle(module_manage->resource[moduleId].module_extension);
    	}
    	return MV_FALSE;
}

MV_BOOLEAN	OSD_IsReset(MV_PVOID extension)
{
	POSD_Extension pOSD = (POSD_Extension)extension;

	PModule_Manage module_manage = &pOSD->Module_Manage;
	
	MV_U16 moduleId = Module_GetModuleId(extension);

    	while(module_set[++moduleId].is_hardware_reseting)
    	{
		MV_ASSERT(module_manage->resource[moduleId].module_extension);
    		return module_set[moduleId].is_hardware_reseting(module_manage->resource[moduleId].module_extension);
    	}

    	return MV_FALSE;
}

void	OSD_ResetBus(MV_PVOID extension)
	
{
	POSD_Extension pOSD = (POSD_Extension)extension;
	
	PModule_Manage module_manage = &pOSD->Module_Manage;
	
	MV_U16 moduleId = Module_GetModuleId(extension);

	while(module_set[++moduleId].reset)
    	{
    		module_set[moduleId].reset(module_manage->resource[moduleId].module_extension);
    		break;
    	}
}
#endif

/*
 * 
 * Other exposed functions
 *
 */
extern void OSD_HandleWaitingList(POSD_Extension pOSD);
/* The extension is the calling module extension. It can be any module extension. */
void OSD_ModuleStarted(MV_PVOID extension)
{
	POSD_Extension pOSD = (POSD_Extension)Module_GetOSDExtension(extension);
	PModule_Manage p_module_manage = &pOSD->Module_Manage;
	MV_U16 module_id = Module_GetModuleId(extension);

	/*MV_ASSERT( (module_id>=0) && (module_id<MAX_MODULE_NUMBER) );*/
	MV_ASSERT( module_id<MAX_MODULE_NUMBER );

	p_module_manage->status |= (1<<module_id);

	/* Whether all the modules are started. */
	if ( module_id==0 )
	{
		MV_PRINT("success to init chip\n");
		
		up(&pOSD->sem);
		/* We are totally ready for requests handling. */
		pOSD->State = DRIVER_STATUS_STARTED;

		/* Trigger request handling */
		OSD_HandleWaitingList(pOSD);

		/* Module 0 is the last module */
		OSD_ModuleNotification(pOSD, EVENT_MODULE_ALL_STARTED, 0);
	}
	else
	{
		/* Start the next module. From the lowerer to the higher. */
		Module_StartAll(p_module_manage, module_id-1);
	}
}

void OSD_GetResource(
	MV_PVOID extension,
	enum Resource_Type type,
	MV_PVOID resource
	)
{
	POSD_Extension pOSD = (POSD_Extension)Module_GetOSDExtension(extension);
	PModule_Manage pModuleManage = &pOSD->Module_Manage;
	MV_U16 moduleId = Module_GetModuleId(extension);

	PAssigned_Uncached_Memory pResource = (PAssigned_Uncached_Memory)resource;

	if ( type==RESOURCE_UNCACHED_MEMORY )
	{
		pResource->Physical_Address.value = pModuleManage->resource[moduleId].uncached_physical_address.value;
		pResource->Virtual_Address = pModuleManage->resource[moduleId].uncached_address;
		pResource->Byte_Size = pModuleManage->resource[moduleId].uncached_size;
		return;
	}

	MV_ASSERT(MV_FALSE);
}

void OSD_GetControllerInfor(
	IN MV_PVOID extension,
	OUT PController_Infor pController
	)
{
	POSD_Extension pOSD = (POSD_Extension)Module_GetOSDExtension(extension);
	MV_U8 i;
	for ( i=0; i<MAX_BASE_ADDRESS; i++ )
	{
		pController->Base_Address[i] = pOSD->Base_Address[i];
	}
	pController->Vendor_Id = pOSD->Vendor_Id;
	pController->Device_Id = pOSD->Device_Id;
	pController->Revision_Id = pOSD->Revision_Id;
}

void OSD_SleepMillisecond(
	IN MV_PVOID extension,
	IN MV_U32 millisecond
	)
{
	mdelay(millisecond);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) && defined(__x86_64__)
	touch_nmi_watchdog();

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	/* can't remember exactly in what version this was introduced. */
	touch_softlockup_watchdog();
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11) */

#endif

}
void OSD_SleepMicrosecond(
	IN MV_PVOID extension,
	IN MV_U32 microseconds
	)
{
	while (microseconds > 1000) {
		udelay(1000);
		microseconds -= 1000;
	}

	udelay(microseconds);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) && defined(__x86_64__)
	touch_nmi_watchdog();

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	/* can't remember exactly in what version this was introduced. */
	touch_softlockup_watchdog();
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11) */

#endif
								
}

void OSD_TimerRoutine(unsigned long DeviceExtension) //TO DO ???
{
#ifndef SUPPORT_TIMER
	POSD_Extension pOSD   = (POSD_Extension)Module_GetOSDExtension(DeviceExtension);
	PTimer_Module  pTimer = &pOSD->Timer_Module;
	unsigned long  flags;

#ifdef __AC_DBG__
	unsigned long now;
	
	MV_DASSERT( pTimer->routine!=NULL );
	
	now = jiffies;
	spin_lock_irqsave(pOSD->lock, flags);
	pTimer->routine(pTimer->context);
	spin_unlock_irqrestore(pOSD->lock, flags);
	MV_DBG(DMSG_ACDB, "Timer routine %p used %lu jiffies.\n", 
	       pTimer->routine, jiffies-now);
	/*dump_stack()*/
#else /* __AC_DBG__ */
	MV_DASSERT( pTimer->routine!=NULL );
	spin_lock_irqsave(pOSD->lock, flags);
	pTimer->routine(pTimer->context);
	spin_unlock_irqrestore(pOSD->lock, flags);
#endif /* __AC_DBG__ */
#endif /* SUPPORT_TIMER */

}

void OSD_RequestTimer(
	IN MV_PVOID extension,
	IN MV_U32 millisecond,
	MV_VOID (*routine) (MV_PVOID)
	)
{
	POSD_Extension pOSD = (POSD_Extension)Module_GetOSDExtension(extension);
	PTimer_Module pTimer = &pOSD->Timer_Module;
	
	pTimer->routine = routine;
	pTimer->context = extension;

	del_timer(&pOSD->timer);
	pOSD->timer.function = OSD_TimerRoutine;
	pOSD->timer.data = (unsigned long)extension;
	pOSD->timer.expires = jiffies + 1 + (millisecond * HZ) / 1000;
	add_timer(&pOSD->timer);
}

MV_VOID OSD_ModuleBist(MV_PVOID extension)
{
	POSD_Extension pOSD = (POSD_Extension)extension;
	MV_PRINT("OSD: Io_Count=0x%x.\n", pOSD->Io_Count);
}

MV_VOID 
OSD_GetNextModuleSubmitFunction(
	IN MV_PVOID self_extension,
	OUT MV_PVOID *next_extension,
	OUT MV_VOID (**next_function)(MV_PVOID , PMV_Request)
	)
{
	POSD_Extension pOSD = (POSD_Extension)Module_GetOSDExtension(self_extension);
	MV_U16 module_id = Module_GetModuleId(self_extension);

	module_id++;
	*next_extension = pOSD->Module_Manage.resource[module_id].module_extension;
	*next_function = module_set[module_id].module_submitrequest;
}

MV_VOID 
OSD_GetNextModuleExtension(
	IN MV_PVOID self_extension,
	OUT MV_PVOID *next_extension
	)
{
	POSD_Extension pOSD = (POSD_Extension)Module_GetOSDExtension(self_extension);
	MV_U16 module_id = Module_GetModuleId(self_extension);

	module_id++;
	*next_extension = pOSD->Module_Manage.resource[module_id].module_extension;
}

MV_PVOID
OSD_GetModuleExtension(
   IN MV_PVOID self_extension,
   IN MV_U8 module_id
   )
{
	POSD_Extension pOSD = (POSD_Extension)Module_GetOSDExtension(self_extension);

	MV_DASSERT( module_id<MAX_MODULE_NUMBER );
	return pOSD->Module_Manage.resource[module_id].module_extension;
}

/* Seconds since midnight, Jan 1, 1970 */
MV_U32 OSD_GetTimeInSecond(void)
{
	/*
	 * Seconds since January 1, 1970, 00:00:00 GMT. 
	 * A negative number is the number of milliseconds before January 1, 1970, 00:00:00 GMT.
	 */
	struct timeval tv;
	do_gettimeofday(&tv);
	return (MV_U32)tv.tv_sec;
}

/* Millisecond passed in this day */
MV_U32 OSD_GetMillisecondInDay(void)
{
	MV_U32 ret = 0;
	struct timespec tv;
	struct timeval x;
	do_gettimeofday(&x);
	tv.tv_sec = x.tv_sec;
	tv.tv_nsec = x.tv_usec*NSEC_PER_SEC;
	ret = (MV_U32)(((signed long long) tv.tv_sec * NSEC_PER_SEC) + tv.tv_nsec);
	return ret;
}

void osd_spin_lock_irq(spinlock_t* plock)
{
	WARN_ON(irqs_disabled());
	spin_lock_irq(plock);                             	
}

void osd_spin_unlock_irq(spinlock_t* plock)
{
	spin_unlock_irq(plock);                             	
}

#ifdef SUPPORT_EVENT
MV_BOOLEAN OSD_AddEvent( 
    IN MV_PVOID extension,
    IN MV_U32 eventID,
    IN MV_U16 deviceID,        
    IN MV_U8 severityLevel,        
    IN MV_U8 param_cnt, 
    IN MV_PU32 params
    )
{
	POSD_Extension pOSD = (POSD_Extension)Module_GetOSDExtension(extension);
	PDriver_Event_Entry pEvent;
	static MV_U32 sequenceNo = 1;
	if (param_cnt > MAX_EVENT_PARAMS)
		return MV_FALSE;

	if ( List_Empty(&pOSD->Free_Events) )
	{
		// No free entry, we need to reuse the oldest entry from Stored_Events.
		MV_ASSERT(!List_Empty(&pOSD->Stored_Events));
		MV_ASSERT(pOSD->Num_Stored_Events == MAX_EVENTS);
		pEvent = List_GetFirstEntry((&pOSD->Stored_Events), Driver_Event_Entry, Queue_Pointer);
	}
	else
	{
		pEvent = List_GetFirstEntry((&pOSD->Free_Events), Driver_Event_Entry, Queue_Pointer);
		pOSD->Num_Stored_Events++;
		MV_ASSERT(pOSD->Num_Stored_Events <= MAX_EVENTS);
	}

	pEvent->Event.AdapterID = pOSD->Adapter_Device_Number;  
	pEvent->Event.EventID = eventID; 
	pEvent->Event.SequenceNo = sequenceNo++;
	pEvent->Event.Severity = severityLevel;
	pEvent->Event.DeviceID = deviceID;
//	pEvent->Event.Param_Cnt = param_cnt;
	pEvent->Event.TimeStamp = OSD_GetTimeInSecond();

	if (param_cnt > 0 && params != NULL)
		MV_CopyMemory( (MV_PVOID)pEvent->Event.Params, (MV_PVOID)params, param_cnt * 4 );

	List_AddTail( &pEvent->Queue_Pointer, &pOSD->Stored_Events );

	return MV_TRUE;
}
#endif
