#if !defined(OSD_EXPOSE_H)
#define OSD_EXPOSE_H

#ifdef SUPPORT_EVENT
#include "com_event_struct.h"
#include "com_event_define.h"
#endif

/*
 * Module_Interface function table
 */
MV_U32 OSD_ModuleGetResourceRequirement(enum Resource_Type type, MV_U16 maxIo);
void OSD_ModuleInitialize(MV_PVOID, MV_U32, MV_U16);
void OSD_ModuleStart(MV_PVOID);
void OSD_ModuleStop(MV_PVOID);
void OSD_ModuleNotification(MV_PVOID, enum Module_Event, MV_U32);
void OSD_ModuleSubmitRequest(MV_PVOID, PMV_Request);
void OSD_ModuleBist(MV_PVOID);
void OSD_ModuleReset(MV_PVOID extension);

#ifdef USE_MODULE_CO_RESET
MV_BOOLEAN OSD_WaitingForIdle(MV_PVOID extension);
MV_BOOLEAN	OSD_IsReset(MV_PVOID extension);
void	OSD_ResetBus(MV_PVOID extension);
#endif

/*
 * Other exposed functions
 */
void OSD_ModuleStarted(MV_PVOID extension);

void OSD_GetResource(
	MV_PVOID extension,
	enum Resource_Type type,
	MV_PVOID resource
	);
/* 
 * For OSD_GetResource. If the type is RESOURCE_UNCACHED_MEMORY, 
 * resource data type is PAssigned_Uncached_Memory.
 */
typedef struct _Assigned_Uncached_Memory
{
	MV_PVOID			Virtual_Address;
	MV_PHYSICAL_ADDR	Physical_Address;
	MV_U32				Byte_Size;
	MV_U32				Reserved0;
} Assigned_Uncached_Memory, *PAssigned_Uncached_Memory;

typedef struct _Controller_Infor
{
	MV_LPVOID Base_Address[MAX_BASE_ADDRESS];
	MV_U16 Vendor_Id;
	MV_U16 Device_Id;
	MV_U8 Revision_Id;
	MV_U8 Reserved[3];
} Controller_Infor, *PController_Infor;

#ifdef SUPPORT_EVENT
// wrapper for DriverEvent, needed to implement queue
typedef struct _Driver_Event_Entry
{
	List_Head Queue_Pointer;
	DriverEvent Event;
} Driver_Event_Entry, *PDriver_Event_Entry;
#endif

void OSD_GetControllerInfor(
	IN MV_PVOID extension,
	OUT PController_Infor pController
	);

void OSD_SleepMillisecond(
	IN MV_PVOID extension,
	IN MV_U32 millisecond
	);

void OSD_SleepMicrosecond(
	IN MV_PVOID extension,
	IN MV_U32 microsecond
	);

void OSD_RequestTimer(
	IN MV_PVOID extension,
	IN MV_U32 millisecond,
	MV_VOID (*TimerService) (MV_PVOID)
	);

void OSD_GetNextModuleSubmitFunction(
	IN MV_PVOID self_extension,
	OUT MV_PVOID *next_extension,
	OUT MV_VOID (**next_function)(MV_PVOID , PMV_Request)
	);

MV_PVOID
OSD_GetModuleExtension(
   IN MV_PVOID self_extension,
   IN MV_U8 module_id
   );

/* -------- Added by xxp -------- */
void OSD_GetNextModuleExtension(
	IN MV_PVOID self_extension,
	OUT MV_PVOID *next_extension
	);
/* -------- End -------- */

/* It returns the seconds since midnight, Jan 1, 1970*/
MV_U32 OSD_GetTimeInSecond(void);

/* It returns the millisecond since midnight. It's in one day only. */
MV_U32 OSD_GetMillisecondInDay(void);

#ifdef SUPPORT_EVENT
MV_BOOLEAN OSD_AddEvent( 
    IN MV_PVOID extension,
    IN MV_U32 eventId,
    IN MV_U16 deviceId,        
    IN MV_U8 severityLevel,        
    IN MV_U8 param_cnt, 
    IN MV_PU32 params
    );
#endif

void mvGetAdapterInfo( MV_PVOID This, PMV_Request pReq );

void osd_spin_lock_irq(spinlock_t* plock);
void osd_spin_unlock_irq(spinlock_t* plock);

#ifdef SUPPORT_EVENT
void mvGetEvent( MV_PVOID This, PMV_Request pReq );
#endif

#endif /* OSD_EXPOSE_H */
