#include "mv_include.h"
#include "linux_main.h"
#include "osd_mod.h"

#include "osd_inter.h"

/*
 * Pre-defined module function table
 * Please match this predefined module set with the enum Module_Id
 * Pay attention: If you want to change the following data structure,
 * please change Module_Id as well.
 */
Module_Interface module_set[MAX_MODULE_NUMBER] = 
{
	{
		MODULE_OSD,
		OSD_ModuleGetResourceRequirement,
		OSD_ModuleInitialize,
		OSD_ModuleStart,
		OSD_ModuleStop,
		NULL, /* - USE ME AND DIE!! OSD_ModuleNotification, */
		OSD_ModuleSubmitRequest,
		OSD_ModuleReset,
		OSD_ModuleBist
	},
#ifdef RAID_DRIVER
	{
		MODULE_RAID,
		RAID_ModuleGetResourceRequirement,
		RAID_ModuleInitialize,
		RAID_ModuleStart,
		RAID_ModuleStop,
		RAID_ModuleNotification,
		RAID_ModuleSubmitRequest,
		RAID_ModuleReset,
		RAID_ModuleBist
	},
#endif
	{
		MODULE_CORE,
		Core_ModuleGetResourceRequirement,
		Core_ModuleInitialize,
		Core_ModuleStart,
		Core_ModuleStop,
		Core_ModuleNotification,
		Core_ModuleSubmitRequest,
		Core_ModuleReset,
		Core_ModuleBist
	}
};

void Module_InitializeAll(PModule_Manage p_module_manage, MV_U16 max_io)
{
	MV_I8 i = 0;
	MV_PVOID module_extension = NULL;
	MV_U32 extension_size = 0;

	/* Module initialization is one synchronized function. */
	for ( i=MAX_MODULE_NUMBER-1; i>=0; i-- )
	{
		/* I use this chance to check whether the module_set matches with Module_Id */
		MV_ASSERT( module_set[i].module_id==i );

		if ( module_set[i].module_initialize )
		{
			module_extension = p_module_manage->resource[i].module_extension;
			extension_size = p_module_manage->resource[i].extension_size;
			module_set[i].module_initialize(module_extension, extension_size, max_io);
		}
	}
}

void Module_StartAll(PModule_Manage p_module_manage, MV_U8 begin_module)
{
	MV_I8 i = 0;

        /* 
         * Start module from the lower level, the first one is the core driver.
         * Every time we only start one module.
         */
	for ( i=begin_module; i>=0; i-- )
	{
		MV_ASSERT(begin_module<MAX_MODULE_NUMBER);
		if ( module_set[i].module_start )
		{
			module_set[i].module_start(
				p_module_manage->resource[i].module_extension);
			return;
		}

		/* If the module_start function is NULL, continue to the next. */
		p_module_manage->status |= (1<<i);
	}
}

MV_U32 mod_get_mem_size(POSD_Extension pOSD, enum Resource_Type type, 
			MV_U16 max_io)
{
	int i = 0;
	unsigned long requirement = 0;
	unsigned long  oneRequire = 0;

	for (i=0; i<MAX_MODULE_NUMBER; i++) {
		if (module_set[i].get_mem_size != NULL) {
			oneRequire = module_set[i].get_mem_size(type, max_io);
			requirement += ROUNDING(oneRequire, 8);
			MV_DBG(DMSG_KERN,
			       "%s requirement for module %d is 0x%lx.\n", 
			       type == RESOURCE_CACHED_MEMORY? "Cached memory" : "Uncached memory",
			       i, 
			       oneRequire);
			if (type == RESOURCE_UNCACHED_MEMORY) {
				MV_PVOID uncached_virtual = NULL;
				uncached_virtual = kmalloc(oneRequire, 
							   GFP_KERNEL);
				if (uncached_virtual != NULL)
					pOSD->uncached_virtual_address[i] = uncached_virtual;
				else
					return -1;
			}
		}
	}

	/* Each extension needs one extension header which is hidden from module. */
	if ( type==RESOURCE_CACHED_MEMORY )
		requirement += MODULE_HEADER_SIZE * MAX_MODULE_NUMBER;

	MV_DBG(DMSG_KERN, "%s requirement totally is 0x%lx.\n",
	       type==RESOURCE_CACHED_MEMORY? "Cached memory" : "Uncached memory",
	       requirement);

	return requirement;
}

void Module_AssignModuleExtension(MV_PVOID device_extension, MV_U16 max_io)
{
	MV_PTR_INTEGER ptemp = (MV_PTR_INTEGER)device_extension;
	POSD_Extension pOSD = NULL;
	PModule_Manage module_manage = NULL;
	PModule_Header header = NULL;
	MV_U8 module_id;
	MV_U32 require;

	MV_ASSERT(MODULE_OSD==0);
	pOSD = (POSD_Extension)( (MV_PTR_INTEGER)device_extension+MODULE_HEADER_SIZE );
	module_manage = &pOSD->Module_Manage;

	for (module_id=0; module_id<MAX_MODULE_NUMBER; module_id++) {
		if (module_set[module_id].get_mem_size==NULL )
			continue;

		require = module_set[module_id].get_mem_size(RESOURCE_CACHED_MEMORY, max_io);
		require = ROUNDING(require, 8);
		
		header = (PModule_Header)ptemp;
		header->extension_size = require;
		header->header_size = MODULE_HEADER_SIZE;
		header->module_id = module_id;
		header->osd_extension = pOSD;

		module_manage->resource[module_id].module_extension = (MV_PVOID)(ptemp+MODULE_HEADER_SIZE);
		module_manage->resource[module_id].extension_size = require;

		ptemp += MODULE_HEADER_SIZE+require;
	}
}

void
Module_AssignUncachedMemory(
	IN PModule_Manage module_manage,
	IN MV_PVOID virtual_addr,
	IN MV_PHYSICAL_ADDR physical_addr,
	IN MV_U32 memory_size,
	IN MV_U16 max_io,
	MV_U8 module_id
	)
{
	MV_PTR_INTEGER temp_virtual = (MV_PTR_INTEGER)virtual_addr;
	//MV_PHYSICAL_ADDR temp_physical = physical_addr;

	MV_U32 require;

	/* Assign Uncached Memory */
	if ( module_set[module_id].get_mem_size == NULL )
		return;

	require = module_set[module_id].get_mem_size(RESOURCE_UNCACHED_MEMORY,
						     max_io);
	require = ROUNDING(require, 8);	

	module_manage->resource[module_id].uncached_size = require;
	module_manage->resource[module_id].uncached_address = (MV_PVOID)virtual_addr;
	module_manage->resource[module_id].uncached_physical_address = physical_addr;

	temp_virtual += require;
	/* Do we have enough uncached memory? */
	MV_ASSERT( (temp_virtual-(MV_PTR_INTEGER)virtual_addr)<=memory_size );
}

void Module_StopAll(PModule_Manage p_module_manage)
{
	MV_I8 i = 0;
	MV_PVOID module_extension = NULL;

	/* Module stop is one synchronized function. */
	for ( i=MAX_MODULE_NUMBER-1; i>=0; i-- )
	{
		if ( module_set[i].module_stop )
		{
			module_extension = p_module_manage->resource[i].module_extension;
			module_set[i].module_stop(module_extension);
		}
	}
}
